from __future__ import annotations

__all__ = (
    "Middleware",
    "chat_action",
)

import asyncio
from typing import TYPE_CHECKING
from typing import Any
from typing import Generic
from typing import TypeVar

if TYPE_CHECKING:
    from collections.abc import Callable

    from .enums import ChatAction
    from .types import Message
    from .typing import HandlerCallableType
    from .typing import HandlerMiddleware

T = TypeVar("T")


class Middleware(Generic[T]):

    def __init__(
        self,
        obj: Callable[[T], T],
        *args: Any,
        **kwargs: Any,
    ) -> None:
        self.obj = obj
        self.args = args
        self.kwargs = kwargs

    def __call__(self, obj: T) -> T:
        return self.obj(obj, *self.args, **self.kwargs)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Middleware) and (  # yapf: disable
            other is self or (
                other.obj is self.obj
                and other.args == self.args
                and other.kwargs == self.kwargs
            )
        )


def chat_action(action: ChatAction) -> HandlerMiddleware[Message]:
    """Send chat action and automatically update it if your operation takes
    more than 5 seconds to complete.

    Usage::

        @on_message(middleware=[chat_action(ChatAction.UPLOAD_PHOTO)])
        async def slow_operation(message: Message) -> None:
            ...

    See: :class:`ChatAction <yatbaf.enums.ChatAction>`

    :param action: Chat action.
    """

    def outer(
        handler: HandlerCallableType[Message]
    ) -> HandlerCallableType[Message]:

        async def inner(update: Message) -> None:
            event = asyncio.Event()

            async def _task() -> None:
                while not event.is_set():
                    await update.bot.send_chat_action(update.chat.id, action)
                    await asyncio.sleep(4.5)  # update status every 4.5 sec

            _ = asyncio.create_task(
                _task(), name=f"chat-action-{update.message_id}"
            )
            try:
                await handler(update)
            finally:
                event.set()

        return inner

    return outer



import pytest

from yatbaf.filters import User

USER_ID = 1212
USER_USERNAME = "test_user"


@pytest.fixture(autouse=True)
def _set_user_attrs(user):
    user.id = USER_ID
    user.username = USER_USERNAME


def test_priority():
    assert User(123).priority == {"sender": (1, 100)}
    assert User(123, 321).priority == {"sender": (2, 100)}


@pytest.mark.asyncio
@pytest.mark.parametrize("username", ("@test_user", "test_user", USER_ID))
async def test_true(message, username):
    assert await User(username).check(message)


@pytest.mark.asyncio
@pytest.mark.parametrize("username", ("@testuser", "test_user1", 23345678))
async def test_false(message, username):
    assert not await User(username).check(message)


@pytest.mark.asyncio
@pytest.mark.parametrize("username", ("@test_user", "test_user", USER_ID))
async def test_invert(message, username):
    assert not await (~User(username)).check(message)



import unittest.mock as mock

import pytest

from yatbaf.enums import UpdateType
from yatbaf.filters import Chat
from yatbaf.handler import Handler


def test_new_hander(handler_fn):
    handler = Handler(handler_fn, update_type=UpdateType.MESSAGE)
    assert not handler._filters
    assert not handler._middleware
    assert handler._update_type is UpdateType.MESSAGE
    assert str(handler) == f"<Handler[type=message,id=0x{id(handler):x}]>"
    assert handler.fn is handler_fn


def test_on_registration(monkeypatch, handler_fn):
    monkeypatch.setattr(Handler, "_resolve_guards", rg := mock.Mock())
    monkeypatch.setattr(Handler, "_resolve_dependencies", rd := mock.Mock())
    monkeypatch.setattr(Handler, "_resolve_middleware", rm := mock.Mock())
    handler = Handler(handler_fn, "message")
    handler.on_registration()
    rg.assert_called_once()
    rm.assert_called_once()
    rd.assert_called_once()


async def fn(_):
    pass


def middleware(handler):

    async def wrapper(update):
        await handler(update)

    return wrapper


@pytest.mark.parametrize(
    "objs",
    [
        (
            Handler(fn=fn, update_type=UpdateType.MESSAGE),
            Handler(fn=fn, update_type=UpdateType.MESSAGE),
        ),
        (
            Handler(
                fn=fn,
                update_type=UpdateType.MESSAGE,
            ),
            Handler(
                fn=fn,
                update_type=UpdateType.MESSAGE,
                filters=[Chat("group")],
            ),
        ),
        (
            Handler(
                fn=fn,
                update_type=UpdateType.MESSAGE,
                middleware=[middleware],
            ),
            Handler(
                fn=fn,
                update_type=UpdateType.MESSAGE,
                middleware=[middleware],
            ),
        ),
    ]
)
def test_eq(objs):
    handler1, handler2 = objs
    assert handler1 == handler2


async def fn2(_):
    pass


@pytest.mark.parametrize("h1", [Handler(fn, update_type=UpdateType.MESSAGE)])
@pytest.mark.parametrize(
    "h2",
    [
        Handler(fn, update_type=UpdateType.EDITED_MESSAGE),
        Handler(fn, update_type=UpdateType.MESSAGE, middleware=[object()]),
        Handler(fn2, update_type=UpdateType.MESSAGE),
    ]
)
def test_not_eq(h1, h2):
    assert h1 != h2



from yatbaf.utils import wrap_middleware


def create_middleware(lst: list, m: int):

    def middleware(fn):

        def wrapper():
            lst.append(m)
            fn()

        return wrapper

    return middleware


def test_wrap():
    list_ = []
    middleware = [create_middleware(list_, i) for i in range(4)]

    def target():
        list_.append("fn")

    result = wrap_middleware(target, middleware)
    result()
    assert list_ == [0, 1, 2, 3, "fn"]



from __future__ import annotations

__all__ = (
    "BaseFilter",
    "And",
    "Or",
    "Not",
)

from abc import ABC
from abc import abstractmethod
from collections import defaultdict
from typing import TYPE_CHECKING
from typing import Generic
from typing import final

from yatbaf.exceptions import FilterCompatError
from yatbaf.typing import UpdateT

if TYPE_CHECKING:
    from collections.abc import Sequence

    from yatbaf.typing import FilterPriority


class BaseFilter(ABC, Generic[UpdateT]):
    """Base class for filters."""

    @final
    def __or__(self, other: object) -> BaseFilter[UpdateT]:
        if not isinstance(other, BaseFilter):
            return NotImplemented
        return Or(self, other)

    @final
    def __and__(self, other: object) -> BaseFilter[UpdateT]:
        if not isinstance(other, BaseFilter):
            return NotImplemented
        return And(self, other)

    @final
    def __invert__(self) -> BaseFilter[UpdateT]:
        return Not(self)

    @final
    @classmethod
    def incompat(cls, filter: type[BaseFilter]) -> None:
        _conflicts_map[cls].add(filter)
        _conflicts_map[filter].add(cls)

    @property
    @abstractmethod
    def priority(self) -> FilterPriority:
        pass

    @abstractmethod
    async def check(self, update: UpdateT) -> bool:
        """Check filter"""


@final
class And(BaseFilter[UpdateT]):

    def __init__(
        self, f1: BaseFilter[UpdateT], f2: BaseFilter[UpdateT]
    ) -> None:
        self.f1 = f1
        self.f2 = f2

    @property
    def priority(self) -> FilterPriority:
        return merge_priority(self.f1.priority, self.f2.priority)

    async def check(self, update: UpdateT) -> bool:
        return await self.f1.check(update) and await self.f2.check(update)


@final
class Or(BaseFilter[UpdateT]):

    def __init__(
        self, f1: BaseFilter[UpdateT], f2: BaseFilter[UpdateT]
    ) -> None:
        self.f1 = f1
        self.f2 = f2

    @property
    def priority(self) -> FilterPriority:
        return merge_priority(self.f1.priority, self.f2.priority)

    async def check(self, update: UpdateT) -> bool:
        return await self.f1.check(update) or await self.f2.check(update)


@final
class Not(BaseFilter[UpdateT]):

    def __init__(self, f: BaseFilter[UpdateT]) -> None:
        self.f = f

    @property
    def priority(self) -> FilterPriority:
        return self.f.priority

    async def check(self, update: UpdateT) -> bool:
        return not await self.f.check(update)


_conflicts_map: dict[type[BaseFilter], set[type[BaseFilter]]] = defaultdict(set)


def merge_priority(p1: FilterPriority, p2: FilterPriority) -> FilterPriority:
    """:meta private:"""
    result = {**p1}
    for p2_group, p2_prior in p2.items():
        prior: tuple[int, int | tuple[int, int]]
        if p1_prior := result.get(p2_group):
            max_p1, min_p1 = _unpack(p1_prior[1])
            max_p2, min_p2 = _unpack(p2_prior[1])

            prior = (
                p1_prior[0] + p2_prior[0],
                (max([max_p1, max_p2]), min(min_p1, min_p2)),
            )
        else:
            prior = p2_prior

        result[p2_group] = prior
    return result


def _unpack(value: int | tuple[int, int]) -> tuple[int, int]:
    """:meta private:"""
    if isinstance(value, int):
        return value, value
    return value


def is_compat(filter1: BaseFilter, filter2: BaseFilter) -> None:
    """:meta private:"""
    if isinstance(filter1, And | Or):
        for filter in (filter1.f1, filter1.f2):
            is_compat(filter, filter2)

    if isinstance(filter2, And | Or):
        for filter in (filter2.f1, filter2.f2):
            is_compat(filter1, filter)

    if conflicts := _conflicts_map[type(filter1)]:
        if type(filter2) in conflicts:
            raise FilterCompatError(
                f"{type(filter1)} cannot be used with {type(filter2)}."
            )


def check_compatibility(filters: Sequence[BaseFilter], any_: bool) -> None:
    """:meta private:"""
    for i, target in enumerate(filters, start=1):
        if isinstance(target, Not):
            target = target.f

        # check filters inside `And` or `Or` filters
        if isinstance(target, And | Or):
            check_compatibility((target.f1, target.f2), isinstance(target, Or))

        # do not check compat between filters inside the `Or` filter
        if any_:
            continue

        for filter in filters[i:]:
            if isinstance(filter, Not):
                filter = filter.f
            is_compat(target, filter)



from __future__ import annotations

from typing import final

from ..types import chat  # noqa: F401
from .abc import TelegramType


@final
class Story(TelegramType):
    """This object represents a story.

    See: https://core.telegram.org/bots/api#story
    """

    chat: chat.Chat
    """Chat that posted the story."""

    id: int
    """Unique identifier for the story in the chat."""



import pytest

from yatbaf.di import CleanupGroup


@pytest.fixture
def generator(mock_mark):

    async def _generator():
        yield 1
        mock_mark()

    return _generator


@pytest.mark.asyncio
async def test_cleanup(generator, mock_mark):
    cg = CleanupGroup()
    gen = generator()
    await anext(gen)
    cg.add(gen)

    await cg.cleanup()
    mock_mark.assert_called_once()


@pytest.mark.asyncio
async def test_cleanup1(generator, mock_mark):
    cg = CleanupGroup()
    gen1 = generator()
    gen2 = generator()
    await anext(gen1)
    await anext(gen2)
    cg.add(gen1)
    cg.add(gen2)

    await cg.cleanup()
    assert mock_mark.call_count == 2



from __future__ import annotations

__all__ = (
    "Handler",
    "on_message",
    "on_edited_message",
    "on_channel_post",
    "on_business_connection",
    "on_business_message",
    "on_edited_business_message",
    "on_deleted_business_messages",
    "on_edited_channel_post",
    "on_message_reaction",
    "on_message_reaction_count",
    "on_inline_query",
    "on_chosen_inline_result",
    "on_callback_query",
    "on_shipping_query",
    "on_pre_checkout_query",
    "on_poll",
    "on_poll_answer",
    "on_my_chat_member",
    "on_chat_member",
    "on_chat_join_request",
    "on_chat_boost",
    "on_removed_chat_boost",
)

from abc import abstractmethod
from typing import TYPE_CHECKING
from typing import TypeAlias
from typing import final
from typing import overload

from .abc import AbstractHandler
from .di import Provide
from .di import create_dependency_batches
from .di import create_dependency_graph
from .di import get_parameters
from .di import is_reserved_key
from .di import resolve_dependencies
from .di import validate_provider
from .enums import UpdateType
from .exceptions import GuardException
from .filters.base import check_compatibility
from .filters.base import merge_priority
from .typing import UpdateT
from .utils import ensure_unique
from .utils import wrap_middleware

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Iterable
    from collections.abc import Sequence

    from .di import Dependency
    from .filters.base import BaseFilter
    from .types import BusinessConnection
    from .types import BusinessMessagesDeleted
    from .types import CallbackQuery
    from .types import ChatBoostRemoved
    from .types import ChatBoostUpdated
    from .types import ChatJoinRequest
    from .types import ChatMemberUpdated
    from .types import ChosenInlineResult
    from .types import InlineQuery
    from .types import Message
    from .types import MessageReactionCountUpdated
    from .types import MessageReactionUpdated
    from .types import Poll
    from .types import PollAnswer
    from .types import PreCheckoutQuery
    from .types import ShippingQuery
    from .typing import FilterPriority
    from .typing import HandlerCallable
    from .typing import HandlerGuard
    from .typing import HandlerMiddleware


class BaseHandler(AbstractHandler[UpdateT]):

    __slots__ = (
        "parent",
        "_guards",
        "_filters",
        "_middleware",
        "_filters",
        "_priority",
        "_dependencies",
    )

    parent: BaseHandler[UpdateT] | None

    def __init__(
        self,
        filters: Sequence[BaseFilter[UpdateT]] | None = None,
        guards: Sequence[HandlerGuard[UpdateT]] | None = None,
        middleware: Sequence[HandlerMiddleware[UpdateT]] | None = None,
        dependencies: dict[str, Provide] | None = None,
    ) -> None:
        self.parent: BaseHandler[UpdateT] | None = None
        self._filters = tuple(filters or [])
        self._guards = list(guards or [])
        self._middleware = list(middleware or [])
        self._priority = ((0, (0, 0)), (0, (0, 0)), (0, (0, 0)))
        self._dependencies = dependencies or {}

    @final
    def __bool__(self) -> bool:
        return True

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, BaseHandler):
            return NotImplemented

        s_prior = self._priority
        o_prior = other._priority

        # priority (max, min) [content, sender, chat]
        sp = (s_prior[0][1], s_prior[1][1], s_prior[2][1])
        op = (o_prior[0][1], o_prior[1][1], o_prior[2][1])

        # priority max [content, sender, chat]
        sp_max = (sp[0][0], sp[1][0], sp[2][0])
        op_max = (op[0][0], op[1][0], op[2][0])

        # priority min [content, sender, chat]
        sp_min = (sp[0][1], sp[1][1], sp[2][1])
        op_min = (op[0][1], op[1][1], op[2][1])

        # number of filters [chat, sender, content]
        ss = (s_prior[2][0], s_prior[1][0], s_prior[0][0])
        os = (o_prior[2][0], o_prior[1][0], o_prior[0][0])

        if sp_max == op_max:
            if ss == os:
                return sp_min < op_min
            return ss < os
        return sp_max < op_max

    @final
    def __gt__(self, other: object) -> bool:  # noqa: U100
        return NotImplemented

    async def _check_guards(self, update: UpdateT) -> None:
        for guard in self._guards:
            await guard(update)

    async def match(self, update: UpdateT, /) -> bool:
        for filter in self._filters:
            if not await filter.check(update):
                return False
        return True

    def _resolve_middleware(self) -> None:
        self._middleware = ensure_unique(self._middleware)

    def _resolve_guards(self) -> None:
        self._guards = ensure_unique(self._guards)

    def _resolve_filters(self) -> None:
        check_compatibility(self._filters, False)
        self._parse_priority()

    def _parse_priority(self) -> None:
        data: FilterPriority = {}
        for filter in self._filters:
            data = merge_priority(data, filter.priority)

        tmp: list[tuple[int, tuple[int, int]]] = []
        for group in ("content", "sender", "chat"):
            if priority := data.get(group):  # type: ignore[call-overload]
                if isinstance(min_max := priority[1], int):
                    tmp.append((priority[0], (min_max, min_max)))
                else:
                    tmp.append(priority)
            else:
                tmp.append((0, (0, 0)))
        self._priority = tuple(tmp)  # type: ignore[assignment]

    def _get_guards(self) -> Iterable[HandlerGuard[UpdateT]]:
        return self._guards

    def _get_middleware(self) -> Iterable[HandlerMiddleware[UpdateT]]:
        return self._middleware

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        pass

    @property
    @abstractmethod
    def update_type(self) -> UpdateType:
        """Handler type."""

    @abstractmethod
    def on_registration(self) -> None:
        """Init handler."""


@final
class Handler(BaseHandler[UpdateT]):
    """Handler object."""

    __slots__ = (
        "_fn",
        "_update_type",
        "_middleware_stack",
        "_resolved_dependencies",
        "_kwargs",
    )

    def __init__(
        self,
        fn: HandlerCallable[UpdateT],
        update_type: UpdateType | str,
        *,
        filters: Sequence[BaseFilter[UpdateT]] | None = None,
        guards: Sequence[HandlerGuard[UpdateT]] | None = None,
        middleware: Sequence[HandlerMiddleware[UpdateT]] | None = None,
        dependencies: dict[str, Provide] | None = None,
    ) -> None:
        """
        :param fn: Handler callback.
        :param update_type: Handler type. See :class:`~yatbaf.enums.UpdateType`
        :param filters: *Optional.* A sequence of :class:`~yatbaf.filters.base.BaseFilter`.
        :param guards: *Optional.* A sequence of :class:`~yatbaf.typing.HandlerGuard`.
        :param middleware: *Optional.* A sequence of :class:`~yatbaf.typing.HandlerMiddleware`.
        :param dependencies: *Optional.* A mapping of dependency providers.
        """  # noqa: E501
        super().__init__(
            filters=filters,
            guards=guards,
            middleware=middleware,
            dependencies=dependencies,
        )

        self._fn = fn
        self._update_type = UpdateType(update_type)
        self._middleware_stack: HandlerCallable[UpdateT] = self._handle
        self._kwargs = self._get_handler_kwargs(fn)
        self._resolved_dependencies: list[set[Dependency]] = []

    def __repr__(self) -> str:
        return f"<Handler[type={self._update_type!s},id=0x{id(self):x}]>"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Handler) and (  # yapf: disable
            other is self or (
                # type, callback and middleware is the same - handlers are
                # equal. filters don't matter.
                other._update_type == self._update_type
                and other._fn is self._fn
                and other._middleware == self._middleware
            )
        )

    @staticmethod
    def _get_handler_kwargs(fn: HandlerCallable[UpdateT]) -> set[str]:
        return set(get_parameters(fn)[1:])

    @property
    def update_type(self) -> UpdateType:
        """Handler type."""
        return self._update_type

    @property
    def fn(self) -> HandlerCallable[UpdateT]:
        """Original function."""
        return self._fn

    async def handle(self, update: UpdateT, /) -> None:
        await self._middleware_stack(update)

    async def _handle(self, update: UpdateT, /) -> None:
        try:
            if self._guards:
                await self._check_guards(update)
        except GuardException:
            return

        if not self._kwargs:
            await self._fn(update)
            return

        values = {"update": update}
        cg = await resolve_dependencies(self._resolved_dependencies, values)
        kwargs = {k: values[k] for k in self._kwargs}

        try:
            await self._fn(update, **kwargs)
        except Exception as e:
            await cg.throw(e)
            raise

        await cg.cleanup()

    @property
    def _parents(self) -> list[BaseHandler[UpdateT]]:
        result: list[BaseHandler[UpdateT]] = []
        parent: BaseHandler[UpdateT] | None = self
        while parent is not None:
            result.append(parent)
            parent = parent.parent
        return list(reversed(result))

    def _resolve_guards(self) -> None:
        guards: list[HandlerGuard[UpdateT]] = []
        for parent in self._parents:
            guards.extend(parent._get_guards())
        self._guards = ensure_unique(guards)

    def _resolve_middleware(self) -> None:
        middleware: list[HandlerMiddleware[UpdateT]] = []
        for parent in self._parents:
            middleware.extend(parent._get_middleware())
        self._middleware_stack = wrap_middleware(
            self._handle,
            ensure_unique(middleware),  # type: ignore[arg-type]
        )

    def _resolve_dependencies(self) -> None:
        if kwargs := self._kwargs:
            providers = self._get_dependency_providers()
            dependencies = set()
            for key in kwargs:
                if not is_reserved_key(key):
                    dependencies.add(create_dependency_graph(key, providers))
            batches = create_dependency_batches(dependencies)
            self._resolved_dependencies = batches

    def _get_dependency_providers(self) -> dict[str, Provide]:
        providers: dict[str, Provide] = {}
        for parent in self._parents:
            if (d := parent._dependencies) is not None:
                for k, v in d.items():
                    validate_provider(k, v, providers)
                    providers[k] = v
        return providers

    def on_registration(self) -> None:
        self._resolve_guards()
        self._resolve_middleware()
        self._resolve_filters()
        self._resolve_dependencies()


Wrapper: TypeAlias = "Callable[[HandlerCallable[UpdateT]], Handler[UpdateT]]"


@overload
def on_message(__fn: HandlerCallable[Message]) -> Handler[Message]:
    ...


@overload
def on_message(
    *,
    filters: Sequence[BaseFilter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    guards: Sequence[HandlerGuard[Message]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[Message]:
    ...


def on_message(
    __fn: HandlerCallable[Message] | None = None,
    *,
    filters: Sequence[BaseFilter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    guards: Sequence[HandlerGuard[Message]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[Message] | Handler[Message]:
    """Message handler decorator.

    Use this decorator to decorate handler for `message` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[Message]) -> Handler[Message]:
        handler: Handler[Message] = Handler(
            fn=fn,
            update_type=UpdateType.MESSAGE,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_edited_message(__fn: HandlerCallable[Message]) -> Handler[Message]:
    ...


@overload
def on_edited_message(
    *,
    filters: Sequence[BaseFilter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    guards: Sequence[HandlerGuard[Message]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[Message]:
    ...


def on_edited_message(
    __fn: HandlerCallable[Message] | None = None,
    *,
    filters: Sequence[BaseFilter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    guards: Sequence[HandlerGuard[Message]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[Message] | Handler[Message]:
    """Edited message handler decorator.

    Use this decorator to decorate handler for `edited_message` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[Message]) -> Handler[Message]:
        handler: Handler[Message] = Handler(
            fn=fn,
            update_type=UpdateType.EDITED_MESSAGE,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_message_reaction(
    __fn: HandlerCallable[MessageReactionUpdated]
) -> Handler[MessageReactionUpdated]:
    ...


@overload
def on_message_reaction(  # yapf: disable
    *,
    filters: Sequence[BaseFilter[MessageReactionUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[MessageReactionUpdated]] | None = None,  # noqa: E501
    guards: Sequence[HandlerGuard[MessageReactionUpdated]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[MessageReactionUpdated]:
    ...


def on_message_reaction(  # yapf: disable
    __fn: HandlerCallable[MessageReactionUpdated] | None = None,
    *,
    filters: Sequence[BaseFilter[MessageReactionUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[MessageReactionUpdated]] | None = None,  # noqa: E501
    guards: Sequence[HandlerGuard[MessageReactionUpdated]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[MessageReactionUpdated] | Handler[MessageReactionUpdated]:
    """Message reaction handler decorator.

    Use this decorator to decorate handler for `message_reaction` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[MessageReactionUpdated]
    ) -> Handler[MessageReactionUpdated]:
        handler: Handler[MessageReactionUpdated] = Handler(
            fn=fn,
            update_type=UpdateType.MESSAGE_REACTION,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_message_reaction_count(
    __fn: HandlerCallable[MessageReactionCountUpdated]
) -> Handler[MessageReactionCountUpdated]:
    ...


@overload
def on_message_reaction_count(  # yapf: disable
    *,
    filters: Sequence[BaseFilter[MessageReactionCountUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[MessageReactionCountUpdated]] | None = None,  # noqa: E501
    guards: Sequence[HandlerGuard[MessageReactionCountUpdated]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[MessageReactionCountUpdated]:
    ...


def on_message_reaction_count(  # yapf: disable
    __fn: HandlerCallable[MessageReactionCountUpdated] | None = None,
    *,
    filters: Sequence[BaseFilter[MessageReactionCountUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[MessageReactionCountUpdated]] | None = None,  # noqa: E501
    guards: Sequence[HandlerGuard[MessageReactionCountUpdated]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[MessageReactionCountUpdated] | Handler[MessageReactionCountUpdated]:  # noqa: E501
    """Message reaction count handler decorator.

    Use this decorator to decorate handler for `message_reaction` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[MessageReactionCountUpdated]
    ) -> Handler[MessageReactionCountUpdated]:
        handler: Handler[MessageReactionCountUpdated] = Handler(
            fn=fn,
            update_type=UpdateType.MESSAGE_REACTION_COUNT,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_channel_post(__fn: HandlerCallable[Message]) -> Handler[Message]:
    ...


@overload
def on_channel_post(
    *,
    filters: Sequence[BaseFilter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    guards: Sequence[HandlerGuard[Message]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[Message]:
    ...


def on_channel_post(
    __fn: HandlerCallable[Message] | None = None,
    *,
    filters: Sequence[BaseFilter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    guards: Sequence[HandlerGuard[Message]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[Message] | Handler[Message]:
    """Channel post handler decorator.

    Use this decorator to decorate handler for `channel_post` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[Message]) -> Handler[Message]:
        handler: Handler[Message] = Handler(
            fn=fn,
            update_type=UpdateType.CHANNEL_POST,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_edited_channel_post(__fn: HandlerCallable[Message]) -> Handler[Message]:
    ...


@overload
def on_edited_channel_post(
    *,
    filters: Sequence[BaseFilter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    guards: Sequence[HandlerGuard[Message]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[Message]:
    ...


def on_edited_channel_post(
    __fn: HandlerCallable[Message] | None = None,
    *,
    filters: Sequence[BaseFilter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    guards: Sequence[HandlerGuard[Message]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[Message] | Handler[Message]:
    """Edited channel post handler decorator.

    Use this decorator to decorate handler for `edited_channel_post` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[Message]) -> Handler[Message]:
        handler: Handler[Message] = Handler(
            fn=fn,
            update_type=UpdateType.EDITED_CHANNEL_POST,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_inline_query(__fn: HandlerCallable[InlineQuery]) -> Handler[InlineQuery]:
    ...


@overload
def on_inline_query(
    *,
    filters: Sequence[BaseFilter[InlineQuery]] | None = None,
    middleware: Sequence[HandlerMiddleware[InlineQuery]] | None = None,
    guards: Sequence[HandlerGuard[InlineQuery]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[InlineQuery]:
    ...


def on_inline_query(
    __fn: HandlerCallable[InlineQuery] | None = None,
    *,
    filters: Sequence[BaseFilter[InlineQuery]] | None = None,
    middleware: Sequence[HandlerMiddleware[InlineQuery]] | None = None,
    guards: Sequence[HandlerGuard[InlineQuery]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[InlineQuery] | Handler[InlineQuery]:
    """Inline query handler decorator.

    Use this decorator to decorate handler for `inline_query` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[InlineQuery]) -> Handler[InlineQuery]:
        handler: Handler[InlineQuery] = Handler(
            fn=fn,
            update_type=UpdateType.INLINE_QUERY,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_chosen_inline_result(
    __fn: HandlerCallable[ChosenInlineResult]
) -> Handler[ChosenInlineResult]:
    ...


@overload
def on_chosen_inline_result(
    *,
    filters: Sequence[BaseFilter[ChosenInlineResult]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChosenInlineResult]] | None = None,
    guards: Sequence[HandlerGuard[ChosenInlineResult]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[ChosenInlineResult]:
    ...


def on_chosen_inline_result(
    __fn: HandlerCallable[ChosenInlineResult] | None = None,
    *,
    filters: Sequence[BaseFilter[ChosenInlineResult]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChosenInlineResult]] | None = None,
    guards: Sequence[HandlerGuard[ChosenInlineResult]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[ChosenInlineResult] | Handler[ChosenInlineResult]:
    """Chosen inline result handler decorator.

    Use this decorator to decorate handler for `chosen_inline_result` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[ChosenInlineResult]
    ) -> Handler[ChosenInlineResult]:
        handler: Handler[ChosenInlineResult] = Handler(
            fn=fn,
            update_type=UpdateType.CHOSEN_INLINE_RESULT,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_callback_query(
    __fn: HandlerCallable[CallbackQuery]
) -> Handler[CallbackQuery]:
    ...


@overload
def on_callback_query(
    *,
    filters: Sequence[BaseFilter[CallbackQuery]] | None = None,
    middleware: Sequence[HandlerMiddleware[CallbackQuery]] | None = None,
    guards: Sequence[HandlerGuard[CallbackQuery]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[CallbackQuery]:
    ...


def on_callback_query(
    __fn: HandlerCallable[CallbackQuery] | None = None,
    *,
    filters: Sequence[BaseFilter[CallbackQuery]] | None = None,
    middleware: Sequence[HandlerMiddleware[CallbackQuery]] | None = None,
    guards: Sequence[HandlerGuard[CallbackQuery]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[CallbackQuery] | Handler[CallbackQuery]:
    """Callback query handler decorator.

    Use this decorator to decorate handler for `callback_query` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[CallbackQuery]) -> Handler[CallbackQuery]:
        handler: Handler[CallbackQuery] = Handler(
            fn=fn,
            update_type=UpdateType.CALLBACK_QUERY,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_shipping_query(
    __fn: HandlerCallable[ShippingQuery]
) -> Handler[ShippingQuery]:
    ...


@overload
def on_shipping_query(
    *,
    filters: Sequence[BaseFilter[ShippingQuery]] | None = None,
    middleware: Sequence[HandlerMiddleware[ShippingQuery]] | None = None,
    guards: Sequence[HandlerGuard[ShippingQuery]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[ShippingQuery]:
    ...


def on_shipping_query(
    __fn: HandlerCallable[ShippingQuery] | None = None,
    *,
    filters: Sequence[BaseFilter[ShippingQuery]] | None = None,
    middleware: Sequence[HandlerMiddleware[ShippingQuery]] | None = None,
    guards: Sequence[HandlerGuard[ShippingQuery]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[ShippingQuery] | Handler[ShippingQuery]:
    """Shipping query handler decorator.

    Use this decorator to decorate handler for `shipping_query` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[ShippingQuery]) -> Handler[ShippingQuery]:
        handler: Handler[ShippingQuery] = Handler(
            fn=fn,
            update_type=UpdateType.SHIPPING_QUERY,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_pre_checkout_query(
    __fn: HandlerCallable[PreCheckoutQuery]
) -> Handler[PreCheckoutQuery]:
    ...


@overload
def on_pre_checkout_query(
    *,
    filters: Sequence[BaseFilter[PreCheckoutQuery]] | None = None,
    middleware: Sequence[HandlerMiddleware[PreCheckoutQuery]] | None = None,
    guards: Sequence[HandlerGuard[PreCheckoutQuery]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[PreCheckoutQuery]:
    ...


def on_pre_checkout_query(
    __fn: HandlerCallable[PreCheckoutQuery] | None = None,
    *,
    filters: Sequence[BaseFilter[PreCheckoutQuery]] | None = None,
    middleware: Sequence[HandlerMiddleware[PreCheckoutQuery]] | None = None,
    guards: Sequence[HandlerGuard[PreCheckoutQuery]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[PreCheckoutQuery] | Handler[PreCheckoutQuery]:
    """Pre-checkout query handler decorator.

    Use this decorator to decorate handler for `pre_checkout_query` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[PreCheckoutQuery]
    ) -> Handler[PreCheckoutQuery]:
        handler: Handler[PreCheckoutQuery] = Handler(
            fn=fn,
            update_type=UpdateType.PRE_CHECKOUT_QUERY,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_poll(__fn: HandlerCallable[Poll]) -> Handler[Poll]:
    ...


@overload
def on_poll(
    *,
    filters: Sequence[BaseFilter[Poll]] | None = None,
    middleware: Sequence[HandlerMiddleware[Poll]] | None = None,
    guards: Sequence[HandlerGuard[Poll]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[Poll]:
    ...


def on_poll(
    __fn: HandlerCallable[Poll] | None = None,
    *,
    filters: Sequence[BaseFilter[Poll]] | None = None,
    middleware: Sequence[HandlerMiddleware[Poll]] | None = None,
    guards: Sequence[HandlerGuard[Poll]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[Poll] | Handler[Poll]:
    """Poll handler decorator.

    Use this decorator to decorate handler for `poll` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[Poll]) -> Handler[Poll]:
        handler: Handler[Poll] = Handler(
            fn=fn,
            update_type=UpdateType.POLL,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_poll_answer(__fn: HandlerCallable[PollAnswer]) -> Handler[PollAnswer]:
    ...


@overload
def on_poll_answer(
    *,
    filters: Sequence[BaseFilter[PollAnswer]] | None = None,
    middleware: Sequence[HandlerMiddleware[PollAnswer]] | None = None,
    guards: Sequence[HandlerGuard[PollAnswer]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[PollAnswer]:
    ...


def on_poll_answer(
    __fn: HandlerCallable[PollAnswer] | None = None,
    *,
    filters: Sequence[BaseFilter[PollAnswer]] | None = None,
    middleware: Sequence[HandlerMiddleware[PollAnswer]] | None = None,
    guards: Sequence[HandlerGuard[PollAnswer]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[PollAnswer] | Handler[PollAnswer]:
    """Poll answer handler decorator.

    Use this decorator to decorate handler for `poll_answer` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[PollAnswer]) -> Handler[PollAnswer]:
        handler: Handler[PollAnswer] = Handler(
            fn=fn,
            update_type=UpdateType.POLL_ANSWER,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_my_chat_member(
    __fn: HandlerCallable[ChatMemberUpdated]
) -> Handler[ChatMemberUpdated]:
    ...


@overload
def on_my_chat_member(
    *,
    filters: Sequence[BaseFilter[ChatMemberUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatMemberUpdated]] | None = None,
    guards: Sequence[HandlerGuard[ChatMemberUpdated]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[ChatMemberUpdated]:
    ...


def on_my_chat_member(
    __fn: HandlerCallable[ChatMemberUpdated] | None = None,
    *,
    filters: Sequence[BaseFilter[ChatMemberUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatMemberUpdated]] | None = None,
    guards: Sequence[HandlerGuard[ChatMemberUpdated]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[ChatMemberUpdated] | Handler[ChatMemberUpdated]:
    """My chat member handler decorator.

    Use this decorator to decorate handler for `my_chat_member` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[ChatMemberUpdated]
    ) -> Handler[ChatMemberUpdated]:
        handler: Handler[ChatMemberUpdated] = Handler(
            fn=fn,
            update_type=UpdateType.MY_CHAT_MEMBER,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_chat_member(
    __fn: HandlerCallable[ChatMemberUpdated]
) -> Handler[ChatMemberUpdated]:
    ...


@overload
def on_chat_member(
    *,
    filters: Sequence[BaseFilter[ChatMemberUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatMemberUpdated]] | None = None,
    guards: Sequence[HandlerGuard[ChatMemberUpdated]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[ChatMemberUpdated]:
    ...


def on_chat_member(
    __fn: HandlerCallable[ChatMemberUpdated] | None = None,
    *,
    filters: Sequence[BaseFilter[ChatMemberUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatMemberUpdated]] | None = None,
    guards: Sequence[HandlerGuard[ChatMemberUpdated]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[ChatMemberUpdated] | Handler[ChatMemberUpdated]:
    """Chat member handler decorator.

    Use this decorator to decorate handler for `chat_member` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[ChatMemberUpdated]
    ) -> Handler[ChatMemberUpdated]:
        handler: Handler[ChatMemberUpdated] = Handler(
            fn=fn,
            update_type=UpdateType.CHAT_MEMBER,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_chat_join_request(
    __fn: HandlerCallable[ChatJoinRequest]
) -> Handler[ChatJoinRequest]:
    ...


@overload
def on_chat_join_request(
    *,
    filters: Sequence[BaseFilter[ChatJoinRequest]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatJoinRequest]] | None = None,
    guards: Sequence[HandlerGuard[ChatJoinRequest]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[ChatJoinRequest]:
    ...


def on_chat_join_request(
    __fn: HandlerCallable[ChatJoinRequest] | None = None,
    *,
    filters: Sequence[BaseFilter[ChatJoinRequest]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatJoinRequest]] | None = None,
    guards: Sequence[HandlerGuard[ChatJoinRequest]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[ChatJoinRequest] | Handler[ChatJoinRequest]:
    """Chat join request handler decorator.

    Use this decorator to decorate handler for `chat_join_request` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[ChatJoinRequest]
    ) -> Handler[ChatJoinRequest]:
        handler: Handler[ChatJoinRequest] = Handler(
            fn=fn,
            update_type=UpdateType.CHAT_JOIN_REQUEST,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_chat_boost(
    __fn: HandlerCallable[ChatBoostUpdated]
) -> Handler[ChatBoostUpdated]:
    ...


@overload
def on_chat_boost(
    *,
    filters: Sequence[BaseFilter[ChatBoostUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatBoostUpdated]] | None = None,
    guards: Sequence[HandlerGuard[ChatBoostUpdated]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[ChatBoostUpdated]:
    ...


def on_chat_boost(
    __fn: HandlerCallable[ChatBoostUpdated] | None = None,
    *,
    filters: Sequence[BaseFilter[ChatBoostUpdated]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatBoostUpdated]] | None = None,
    guards: Sequence[HandlerGuard[ChatBoostUpdated]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[ChatBoostUpdated] | Handler[ChatBoostUpdated]:
    """Chat boost handler decorator.

    Use this decorator to decorate handler for `chat_boost` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[ChatBoostUpdated]
    ) -> Handler[ChatBoostUpdated]:
        handler: Handler[ChatBoostUpdated] = Handler(
            fn=fn,
            update_type=UpdateType.CHAT_BOOST,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_removed_chat_boost(
    __fn: HandlerCallable[ChatBoostRemoved]
) -> Handler[ChatBoostRemoved]:
    ...


@overload
def on_removed_chat_boost(
    *,
    filters: Sequence[BaseFilter[ChatBoostRemoved]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatBoostRemoved]] | None = None,
    guards: Sequence[HandlerGuard[ChatBoostRemoved]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[ChatBoostRemoved]:
    ...


def on_removed_chat_boost(
    __fn: HandlerCallable[ChatBoostRemoved] | None = None,
    *,
    filters: Sequence[BaseFilter[ChatBoostRemoved]] | None = None,
    middleware: Sequence[HandlerMiddleware[ChatBoostRemoved]] | None = None,
    guards: Sequence[HandlerGuard[ChatBoostRemoved]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[ChatBoostRemoved] | Handler[ChatBoostRemoved]:
    """Removed chat boost handler decorator.

    Use this decorator to decorate handler for `removed_chat_boost` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[ChatBoostRemoved]
    ) -> Handler[ChatBoostRemoved]:
        handler: Handler[ChatBoostRemoved] = Handler(
            fn=fn,
            update_type=UpdateType.REMOVED_CHAT_BOOST,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_business_connection(
    __fn: HandlerCallable[BusinessConnection]
) -> Handler[BusinessConnection]:
    ...


@overload
def on_business_connection(
    *,
    filters: Sequence[BaseFilter[BusinessConnection]] | None = None,
    middleware: Sequence[HandlerMiddleware[BusinessConnection]] | None = None,
    guards: Sequence[HandlerGuard[BusinessConnection]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[BusinessConnection]:
    ...


def on_business_connection(
    __fn: HandlerCallable[BusinessConnection] | None = None,
    *,
    filters: Sequence[BaseFilter[BusinessConnection]] | None = None,
    middleware: Sequence[HandlerMiddleware[BusinessConnection]] | None = None,
    guards: Sequence[HandlerGuard[BusinessConnection]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[BusinessConnection] | Handler[BusinessConnection]:
    """Business connection handler decorator.

    Use this decorator to decorate handler for `business_connection` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(
        fn: HandlerCallable[BusinessConnection]
    ) -> Handler[BusinessConnection]:
        handler: Handler[BusinessConnection] = Handler(
            fn=fn,
            update_type=UpdateType.BUSINESS_CONNECTION,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_business_message(__fn: HandlerCallable[Message]) -> Handler[Message]:
    ...


@overload
def on_business_message(
    *,
    filters: Sequence[BaseFilter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    guards: Sequence[HandlerGuard[Message]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[Message]:
    ...


def on_business_message(
    __fn: HandlerCallable[Message] | None = None,
    *,
    filters: Sequence[BaseFilter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    guards: Sequence[HandlerGuard[Message]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[Message] | Handler[Message]:
    """Business message handler decorator.

    Use this decorator to decorate handler for `business_message` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[Message]) -> Handler[Message]:
        handler: Handler[Message] = Handler(
            fn=fn,
            update_type=UpdateType.BUSINESS_MESSAGE,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_edited_business_message(
    __fn: HandlerCallable[Message]
) -> Handler[Message]:
    ...


@overload
def on_edited_business_message(
    *,
    filters: Sequence[BaseFilter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    guards: Sequence[HandlerGuard[Message]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[Message]:
    ...


def on_edited_business_message(
    __fn: HandlerCallable[Message] | None = None,
    *,
    filters: Sequence[BaseFilter[Message]] | None = None,
    middleware: Sequence[HandlerMiddleware[Message]] | None = None,
    guards: Sequence[HandlerGuard[Message]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[Message] | Handler[Message]:
    """Edited business message handler decorator.

    Use this decorator to decorate handler for `edited_business_message` update.

    See :class:`Handler` for parameters.
    """

    def wrapper(fn: HandlerCallable[Message]) -> Handler[Message]:
        handler: Handler[Message] = Handler(
            fn=fn,
            update_type=UpdateType.EDITED_BUSINESS_MESSAGE,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)


@overload
def on_deleted_business_messages(
    __fn: HandlerCallable[BusinessMessagesDeleted]
) -> Handler[BusinessMessagesDeleted]:
    ...


@overload
def on_deleted_business_messages(  # yapf: disable
    *,
    filters: Sequence[BaseFilter[BusinessMessagesDeleted]] | None = None,
    middleware: Sequence[HandlerMiddleware[BusinessMessagesDeleted]] | None = None,  # noqa: E501
    guards: Sequence[HandlerGuard[BusinessMessagesDeleted]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[BusinessMessagesDeleted]:
    ...


def on_deleted_business_messages(  # yapf: disable
    __fn: HandlerCallable[BusinessMessagesDeleted] | None = None,
    *,
    filters: Sequence[BaseFilter[BusinessMessagesDeleted]] | None = None,
    middleware: Sequence[HandlerMiddleware[BusinessMessagesDeleted]] | None = None,  # noqa: E501
    guards: Sequence[HandlerGuard[BusinessMessagesDeleted]] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Wrapper[BusinessMessagesDeleted] | Handler[BusinessMessagesDeleted]:
    """Deleted business messages handler decorator.

    Use this decorator to decorate handler for `deleted_business_messages` update.

    See :class:`Handler` for parameters.
    """  # noqa: E501

    def wrapper(
        fn: HandlerCallable[BusinessMessagesDeleted]
    ) -> Handler[BusinessMessagesDeleted]:
        handler: Handler[BusinessMessagesDeleted] = Handler(
            fn=fn,
            update_type=UpdateType.DELETED_BUSINESS_MESSAGES,
            filters=filters,
            middleware=middleware,
            guards=guards,
            dependencies=dependencies,
        )
        return handler

    if __fn is None:
        return wrapper
    return wrapper(__fn)



import pytest

from yatbaf.enums import UpdateType
from yatbaf.handler import Handler
from yatbaf.router import OnMessage


def create_middleware(obj):

    def middleware(handler):

        async def wrapper(update):
            update.ctx["test"].append(obj)
            await handler(update)

        return wrapper

    return middleware


def create_callback(obj):

    async def handler(update):
        update.ctx["test"].append(obj)

    return handler


def test_no_middleware(handler_fn):
    handler = Handler(handler_fn, UpdateType.MESSAGE)
    handler._resolve_middleware()
    assert handler._middleware_stack == handler._handle


@pytest.mark.asyncio
async def test_middleware(message):
    message.ctx["test"] = []
    handler = Handler(
        create_callback("h"),
        update_type=UpdateType.MESSAGE,
        middleware=[create_middleware("m")],
    )
    handler._resolve_middleware()
    await handler.handle(message)
    assert message.ctx["test"] == ["m", "h"]


@pytest.mark.asyncio
async def test_middleware_order(message):
    message.ctx["test"] = []
    handler = Handler(
        create_callback("h"),
        update_type=UpdateType.MESSAGE,
        middleware=[
            create_middleware("m1"),
            create_middleware("m2"),
            create_middleware("m3"),
        ],
    )
    handler._resolve_middleware()
    await handler.handle(message)
    assert message.ctx["test"] == ["m1", "m2", "m3", "h"]


@pytest.mark.asyncio
async def test_middleware_order_parent(message):
    message.ctx["test"] = []
    handler = Handler(
        create_callback("h"),
        update_type=UpdateType.MESSAGE,
        middleware=[
            create_middleware("hm1"),
            create_middleware("hm2"),
        ],
    )
    router = OnMessage(
        handler_middleware=[
            create_middleware("rm1"),
            create_middleware("rm2"),
        ],
        routers=[
            OnMessage(
                handlers=[handler],
                handler_middleware=[create_middleware("rm3")],
            )
        ]
    )
    router.on_registration()
    await handler.handle(message)
    assert message.ctx["test"] == ["rm1", "rm2", "rm3", "hm1", "hm2", "h"]


@pytest.mark.asyncio
async def test_middleware_order_parent_local(message):
    message.ctx["test"] = []
    handler = Handler(
        create_callback("h"),
        update_type=UpdateType.MESSAGE,
        middleware=[
            create_middleware("hm1"),
            create_middleware("hm2"),
        ],
    )
    router = OnMessage(
        handler_middleware=[
            create_middleware("rm1"),
            (create_middleware("rm2"), "local"),
        ],
        routers=[
            OnMessage(
                handlers=[handler],
                handler_middleware=[create_middleware("rm3")],
            )
        ]
    )
    router.on_registration()
    await handler.handle(message)
    assert message.ctx["test"] == ["rm1", "rm3", "hm1", "hm2", "h"]


@pytest.mark.asyncio
async def test_middleware_order_parent_duplicate(message):
    middleware = create_middleware("dup")
    message.ctx["test"] = []
    handler = Handler(
        create_callback("h"),
        update_type=UpdateType.MESSAGE,
        middleware=[
            create_middleware("hm1"),
            create_middleware("hm2"),
        ],
    )
    router = OnMessage(
        handler_middleware=[
            create_middleware("rm1"),
            create_middleware("rm2"),
            middleware,
        ],
        routers=[
            OnMessage(
                handlers=[handler],
                handler_middleware=[
                    middleware,
                    create_middleware("rm3"),
                ],
            )
        ]
    )
    router.on_registration()
    await handler.handle(message)
    assert message.ctx["test"] == [
        "rm1", "rm2", "dup", "rm3", "hm1", "hm2", "h"
    ]



import time
import unittest.mock as mock

import pytest

from yatbaf.exceptions import GuardException
from yatbaf.types import Chat
from yatbaf.types import Message
from yatbaf.types import Update
from yatbaf.types import User


@pytest.fixture
def token():
    return "12345678:testtoken"


@pytest.fixture
def user():
    return User(
        id=1010,
        username="testuser",
        is_bot=False,
        first_name="Test",
    )


@pytest.fixture
def chat():
    return Chat(
        id=101010,
        type="group",
        username="testchat",
    )


@pytest.fixture
def message(chat, user):
    return Message(
        from_=user,
        chat=chat,
        date=int(time.time()),
        message_id=101010,
    )


@pytest.fixture
def update(message):
    return Update(
        update_id=9999,
        message=message,
    )


@pytest.fixture
def mock_mark():
    return mock.Mock()


@pytest.fixture
def handler_fn(mock_mark):

    async def fn(update):  # noqa: U100
        mock_mark(update)

    return fn


@pytest.fixture
def guard_true():

    async def fn(_):  # noqa: U101
        return

    return fn


@pytest.fixture
def guard_false():

    async def fn(_):  # noqa: U101
        raise GuardException

    return fn



from yatbaf.middleware import Middleware
from yatbaf.router import OnMessage


def middleware(handler):

    async def wrapper(update):
        await handler(update)

    return wrapper


def test_init_middleware_handler():
    router = OnMessage(handler_middleware=[middleware])
    assert router._handler_middleware == [middleware]
    assert not router._middleware


def test_init_middleware_handler_local():
    router = OnMessage(handler_middleware=[(middleware, "local")])
    assert router._handler_middleware == [(middleware, "local")]
    assert not router._middleware


def test_init_middleware_router():
    router = OnMessage(middleware=[middleware])
    assert router._middleware == [middleware]
    assert not router._handler_middleware


def test_add_middleware_handler():
    router = OnMessage()
    router.add_middleware(middleware)
    assert router._handler_middleware == [middleware]
    assert not router._middleware


def test_add_middleware_handler_local():
    router = OnMessage()
    router.add_middleware(middleware, scope="local")
    assert router._handler_middleware == [(middleware, "local")]
    assert not router._middleware


def test_add_middleware_router():
    router = OnMessage()
    router.add_middleware(middleware, "router")
    assert router._middleware == [middleware]
    assert not router._handler_middleware


def test_middleware_decorator_handler():
    router = OnMessage()

    @router.middleware
    def func(_):
        pass

    assert router._handler_middleware == [func]
    assert not router._middleware


def test_middleware_decorator_handler_local():
    router = OnMessage()

    @router.middleware("local")
    def func(_):
        pass

    assert router._handler_middleware == [(func, "local")]
    assert not router._middleware


def test_middleware_decorator_router():
    router = OnMessage()

    @router.middleware("router")
    def func(_):
        pass

    assert router._middleware == [func]
    assert not router._handler_middleware


def test_middleware_decorator_handler_args():
    router = OnMessage()

    @router.middleware("handler", 1, y=2)
    def func(_, x, *, y):  # noqa: U100
        pass

    assert router._handler_middleware == [Middleware(func, 1, y=2)]


def test_middleware_decorator_handler_local_args():
    router = OnMessage()

    @router.middleware("local", 1, y=2)
    def func(_, x, *, y):  # noqa: U100
        pass

    assert router._handler_middleware == [(Middleware(func, 1, y=2), "local")]


def test_middleware_decorator_duplicate():
    router = OnMessage()

    @router.middleware
    @router.middleware
    def func(_):
        pass

    assert router._handler_middleware == [func, func]


def test_middleware_decorator_router_args():
    router = OnMessage()

    @router.middleware("router", 1)
    def func(_, p):  # noqa: U100,U101
        pass

    assert router._middleware == [Middleware(func, 1)]


def test_middleware_decorator_handler_diff_args():
    router = OnMessage()

    @router.middleware("handler", p=1)
    @router.middleware("hadnler", p=2)
    def func(_, p):  # noqa: U100,U101
        pass

    assert router._handler_middleware == [
        Middleware(func, p=2), Middleware(func, p=1)
    ]



from yatbaf import filters as f
from yatbaf.handler import on_message
from yatbaf.router import OnMessage


def test_fallback():

    @on_message
    async def any_message(_):
        pass

    @on_message(filters=[f.Command("foo"), f.Chat("private")])
    async def foo_private(_):
        pass

    @on_message(filters=[f.Command("foo")])
    async def foo_any(_):
        pass

    router = OnMessage(handlers=[
        any_message,
        foo_any,
        foo_private,
    ])

    assert router._handlers == [
        any_message,
        foo_any,
        foo_private,
    ]

    router.on_registration()
    assert router._handlers == [
        foo_private,
        foo_any,
        any_message,
    ]



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneInt

from .abc import TelegramType


@final
class Birthdate(TelegramType):
    """See: https://core.telegram.org/bots/api#birthdate"""

    day: int
    """Day of the user's birth; 1-31"""

    month: int
    """Month of the user's birth; 1-12"""

    year: NoneInt = None
    """Optional. Year of the user's birth."""



import pytest

from yatbaf.dispatcher import Dispatcher
from yatbaf.handler import on_message


@on_message
async def _handler(update):
    update.ctx["test"].append("h")


handlers = {"message": _handler}


def create_middleware(mark):

    def middleware(handler):

        async def wrapper(update):
            update.event.ctx["test"].append(mark)
            await handler(update)

        return wrapper

    return middleware


@pytest.mark.asyncio
async def test_middleware(update):
    update.event.ctx["test"] = []
    dispatcher = Dispatcher(handlers, middleware=[create_middleware("m")])
    await dispatcher.resolve(update)
    assert update.event.ctx["test"] == ["m", "h"]


@pytest.mark.asyncio
async def test_middleware_order(update):
    update.event.ctx["test"] = []
    dispatcher = Dispatcher(
        handlers,
        middleware=[
            create_middleware(1),
            create_middleware(2),
            create_middleware(3),
        ]
    )
    await dispatcher.resolve(update)
    assert update.event.ctx["test"] == [1, 2, 3, "h"]



import random

import pytest

from yatbaf import filters as f
from yatbaf.handler import on_message


def create_handler(filters=None):

    @on_message(filters=filters)
    async def handler(_):
        pass

    handler.on_registration()
    return handler


fallback = create_handler()

command = create_handler([f.Command("foo")])
command2 = create_handler([f.Command("foo", "bar")])
command_or_text = create_handler([f.Command("foo") | f.Text(startswith="foo")])
command_chat = create_handler([f.Command("foo"), f.Chat("group")])

text = create_handler([f.Text(startswith="hello")])
text2 = create_handler([
    f.Text(
        startswith="hello",
        endswith="world",
    ),
])
text2_any = create_handler([
    f.Text(
        startswith="hello",
        endswith="world",
        any_=True,
    )
])
text3_any = create_handler([
    f.Text(
        startswith="hello",
        endswith="world",
        match="hello-world",
        any_=True,
    )
])

chat = create_handler([f.Chat("private")])
chat2 = create_handler([f.Chat("private", "group")])
chat3 = create_handler([f.Chat("private", "group", "supergroup")])
group_user = create_handler([f.Chat("group"), f.User(123)])

chat_id = create_handler([f.ChatId(1)])
chat_id2 = create_handler([f.ChatId(1, 2)])

content = create_handler([f.Content("text")])
content2 = create_handler([f.Content("text", "document")])
content3 = create_handler([f.Content("text", "document", "audio")])
content_chat = create_handler([f.Content("text"), f.Chat("private")])

user = create_handler([f.User(1)])
user2 = create_handler([f.User(1, 2)])
user3 = create_handler([f.User(1, 2, 3)])
user4 = create_handler([f.User(1, 2, 3, 4)])
not_user = create_handler([~f.User(123)])
user_command = create_handler([f.Command("foo"), f.User("user")])
user_command2 = create_handler([f.Command("foo", "bar"), f.User("user")])
user3_command = create_handler([f.User(1, 2, 3, 4), f.Command("foo")])
user_chat_command = create_handler([
    f.Chat("group"),
    f.User(123),
    f.Command("foo"),
])
user_chat2_command = create_handler([
    f.User(123),
    f.Chat("group") | f.Chat("private"),
    f.Command("foo"),
])
user2_chatid2_command2 = create_handler([
    f.User(1, 2),
    f.ChatId(1, 2),
    f.Command("foo", "bar"),
])


@pytest.mark.parametrize(
    "input_,sorted_",
    [
        [  # 0
            [fallback],
            [fallback],
        ],
        [  # 1
            [
                command,
                command2,
            ],
            [
                command2,
                command,
            ],
        ],
        [  # 2
            [
                command,
                fallback,
                command2,
            ],
            [
                command2,
                command,
                fallback,
            ],
        ],
        [  # 3
            [
                command,
                text,
                text2,
                command_or_text,
            ],
            [
                command_or_text,
                command,
                text2,
                text,
            ],
        ],
        [  # 4
            [
                text,
                command,
                user,
            ],
            [
                command,
                text,
                user,
            ],
        ],
        [  # 5
            [
                text,
                command,
                user,
                chat,
            ],
            [
                command,
                text,
                user,
                chat,
            ],
        ],
        [  # 6
            [
                text,
                command,
                text2,
                chat2,
                chat3,
                user,
                chat,
            ],
            [
                command,
                text2,
                text,
                user,
                chat3,
                chat2,
                chat,
            ],
        ],
        [  # 7
            [
                user,
                user_command,
                user_command2,
                text,
                fallback,
                command,
                command2,
            ],
            [
                user_command2,
                user_command,
                command2,
                command,
                text,
                user,
                fallback,
            ],
        ],
        [  # 8
            [
                user,
                user_command,
                text,
                fallback,
                command,
                command2,
                user_command2,
                text2,
                user2,
                command_or_text,
                user_chat2_command,
            ],
            [
                user_chat2_command,
                user_command2,
                user_command,
                command2,
                command_or_text,
                command,
                text2,
                text,
                user2,
                user,
                fallback,
            ],
        ],
        [  # 9
            [
                user,
                user_command2,
                text,
                fallback,
                user3_command,
                text3_any,
                text2,
                command2,
                chat3,
                chat2,
                user_command,
                chat_id,
                user4,
                chat_id2,
                command,
                group_user,
                command_or_text,
                user_chat_command,
            ],
            [
                user_chat_command,
                user3_command,
                user_command2,
                user_command,
                command2,
                command_or_text,
                command,
                text3_any,
                text2,
                text,
                group_user,
                user4,
                user,
                chat_id2,
                chat_id,
                chat3,
                chat2,
                fallback,
            ],
        ],
        [  # 10
            [
                user,
                user_command,
                user_command2,
                group_user,
                text,
                fallback,
                command,
                command2,
            ],
            [
                user_command2,
                user_command,
                command2,
                command,
                text,
                group_user,
                user,
                fallback,
            ],
        ],
        [  # 11
            [
                chat,
                not_user,
                chat2,
            ],
            [
                not_user,
                chat2,
                chat,
            ],
        ],
        [  # 12
            [
                not_user,
                chat,
                command_or_text,
            ],
            [
                command_or_text,
                not_user,
                chat,
            ],
        ],
        [  # 13
            [
                chat,
                user2_chatid2_command2,
                content,
                text,
                chat2,
                user_command,
                command_or_text,
            ],
            [
                user2_chatid2_command2,
                user_command,
                command_or_text,
                text,
                content,
                chat2,
                chat,
            ],
        ],
        [  # 14
            [
                command2,
                command_chat,
            ],
            [
                command_chat,
                command2,
            ],
        ],
        [  # 15
            [
                command2,
                text2,
                command_chat,
                text3_any,
                user_chat_command,
                user_chat2_command,
                command_or_text,
                group_user,
                not_user,
            ],
            [
                user_chat2_command,
                user_chat_command,
                command_chat,
                command2,
                command_or_text,
                text3_any,
                text2,
                group_user,
                not_user,
            ],
        ],
    ],
)
def test_priority(input_, sorted_):
    random.shuffle(input_)
    assert sorted(input_, reverse=True) == sorted_
    random.shuffle(input_)
    assert sorted(input_, reverse=True) == sorted_



import pytest

from yatbaf.handler import on_message
from yatbaf.router import OnMessage


@pytest.fixture(autouse=True)
def __init_message(message):
    message.ctx["test"] = []


def create_handler(mark):

    @on_message
    async def handler(msg):
        msg.ctx["test"].append(mark)

    return handler


def test_resolve_guards(guard_true, guard_false):
    router = OnMessage(
        guards=[
            guard_true,
            guard_false,
            guard_true,
            guard_true,
        ]
    )
    router._resolve_guards()
    assert router._guards == [guard_true, guard_false]


@pytest.mark.asyncio
async def test_guard_false(message, guard_false):
    router = OnMessage(
        guards=[guard_false],
        handlers=[create_handler(1)],
    )
    router.on_registration()
    await router.handle(message)
    assert message.ctx["test"] == []


@pytest.mark.asyncio
async def test_guard_true(message, guard_true):
    router = OnMessage(
        guards=[guard_true],
        handlers=[create_handler(1)],
    )
    router.on_registration()
    await router.handle(message)
    assert message.ctx["test"] == [1]


@pytest.mark.asyncio
async def test_parent_guard_true(message, guard_true):
    router = OnMessage(
        guards=[guard_true],
        handlers=[create_handler(1)],
        routers=[OnMessage(handlers=[create_handler(2)])],
    )
    router.on_registration()
    await router.handle(message)
    assert message.ctx["test"] == [1]


@pytest.mark.asyncio
async def test_parent_guard_false(message, guard_false):
    router = OnMessage(
        handlers=[create_handler(1)],
        guards=[guard_false],
        routers=[OnMessage(handlers=[create_handler(2)])]
    )
    router.on_registration()
    await router.handle(message)
    assert message.ctx["test"] == []


@pytest.mark.asyncio
async def test_guard_false_both(message, guard_false):
    router = OnMessage(
        guards=[guard_false],
        handlers=[create_handler(1)],
        routers=[
            OnMessage(
                guards=[guard_false],
                handlers=[create_handler(2)],
            ),
        ],
    )
    router.on_registration()
    await router.handle(message)
    assert message.ctx["test"] == []



import pytest

from yatbaf.di import Dependency
from yatbaf.di import Provide
from yatbaf.di import create_dependency_batches


async def dummy() -> None:
    pass


DEPENDENCY_A = Dependency("A", Provide(dummy), [])
DEPENDENCY_B = Dependency("B", Provide(dummy), [])
DEPENDENCY_C1 = Dependency("C1", Provide(dummy), [])
DEPENDENCY_C2 = Dependency("C2", Provide(dummy), [DEPENDENCY_C1])
DEPENDENCY_ALL_EXCEPT_A = Dependency(
    "D", Provide(dummy), [DEPENDENCY_B, DEPENDENCY_C1, DEPENDENCY_C2]
)


@pytest.mark.parametrize(
    "dependency_tree,expected_batches",
    [
        (set(), []),
        ({DEPENDENCY_A}, [{DEPENDENCY_A}]),
        (
            {DEPENDENCY_A, DEPENDENCY_B},
            [
                {DEPENDENCY_A, DEPENDENCY_B},
            ],
        ),
        (
            {DEPENDENCY_C1, DEPENDENCY_C2},
            [
                {DEPENDENCY_C1},
                {DEPENDENCY_C2},
            ],
        ),
        (
            {
                DEPENDENCY_A,
                DEPENDENCY_B,
                DEPENDENCY_C1,
                DEPENDENCY_C2,
                DEPENDENCY_ALL_EXCEPT_A
            },
            [
                {DEPENDENCY_A, DEPENDENCY_B, DEPENDENCY_C1},
                {DEPENDENCY_C2},
                {DEPENDENCY_ALL_EXCEPT_A},
            ],
        ),
        (
            {DEPENDENCY_ALL_EXCEPT_A},
            [
                {DEPENDENCY_B, DEPENDENCY_C1},
                {DEPENDENCY_C2},
                {DEPENDENCY_ALL_EXCEPT_A},
            ],
        ),
    ],
)
def test_dependency_batches(dependency_tree, expected_batches) -> None:
    calculated_batches = create_dependency_batches(dependency_tree)
    assert calculated_batches == expected_batches



import pytest

from yatbaf.exceptions import FrozenInstanceError
from yatbaf.router import OnMessage


async def _guard_func(_):
    return


def test_init_guard_handler():
    router = OnMessage(handler_guards=[_guard_func])
    assert router._handler_guards == [_guard_func]


def test_init_guard_handler_local():
    router = OnMessage(handler_guards=[(_guard_func, "local")])
    assert router._handler_guards == [(_guard_func, "local")]


def test_init_guard_router():
    router = OnMessage(guards=[_guard_func])
    assert router._guards == [_guard_func]


def test_add_guard_handler():
    router = OnMessage()
    router.add_guard(_guard_func)
    assert router._handler_guards == [_guard_func]


def test_add_guard_handler_local():
    router = OnMessage()
    router.add_guard(_guard_func, scope="local")
    assert router._handler_guards == [(_guard_func, "local")]


def test_add_guard_router():
    router = OnMessage()
    router.add_guard(_guard_func, "router")
    assert router._guards == [_guard_func]


def test_add_guard_duplicate():
    router = OnMessage(
        handler_guards=[
            _guard_func,
            _guard_func,
        ],
    )
    assert router._handler_guards == [_guard_func, _guard_func]


def test_guard_decorator_handler():
    router = OnMessage()

    @router.guard
    async def guard(_):
        pass

    assert router._handler_guards == [guard]


def test_guard_decorator_handler_local():
    router = OnMessage()

    @router.guard("local")
    async def guard(_):
        pass

    assert router._handler_guards == [(guard, "local")]


def test_guard_decorator_router():
    router = OnMessage()

    @router.guard("router")
    async def guard(_):
        pass

    assert router._guards == [guard]


def test_frozen():
    router = OnMessage()
    router.on_registration()

    with pytest.raises(FrozenInstanceError):
        router.add_guard(_guard_func)

    with pytest.raises(FrozenInstanceError):

        @router.guard
        async def g_h(_):
            pass

    with pytest.raises(FrozenInstanceError):

        @router.guard("router")
        async def g_r(_):
            pass



from __future__ import annotations

__all__ = (
    "OnMessage",
    "OnEditedMessage",
    "OnChannelPost",
    "OnEditedChannelPost",
    "OnBusinessConnection",
    "OnBusinessMessage",
    "OnEditedBusinessMessage",
    "OnDeletedBusinessMessages",
    "OnMessageReaction",
    "OnMessageReactionCount",
    "OnInlineQuery",
    "OnChosenInlineResult",
    "OnCallbackQuery",
    "OnShippingQuery",
    "OnPreCheckoutQuery",
    "OnPoll",
    "OnPollAnswer",
    "OnMyChatMember",
    "OnChatMember",
    "OnChatJoinRequest",
    "OnChatBoost",
    "OnRemovedChatBoost",
)

from collections import defaultdict
from itertools import count
from typing import TYPE_CHECKING
from typing import Any
from typing import Final
from typing import TypeVar
from typing import cast
from typing import final
from typing import overload

from .enums import UpdateType
from .exceptions import FrozenInstanceError
from .exceptions import GuardException
from .handler import BaseHandler
from .handler import Handler
from .middleware import Middleware
from .types import BusinessConnection
from .types import BusinessMessagesDeleted
from .types import CallbackQuery
from .types import ChatBoostRemoved
from .types import ChatBoostUpdated
from .types import ChatJoinRequest
from .types import ChatMemberUpdated
from .types import ChosenInlineResult
from .types import InlineQuery
from .types import Message
from .types import MessageReactionCountUpdated
from .types import MessageReactionUpdated
from .types import Poll
from .types import PollAnswer
from .types import PreCheckoutQuery
from .types import ShippingQuery
from .typing import UpdateT
from .utils import wrap_middleware
from .warnings import warn_duplicate

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Iterable
    from collections.abc import Iterator
    from collections.abc import Sequence

    from .di import Provide
    from .filters.base import BaseFilter
    from .typing import HandlerCallable
    from .typing import HandlerCallableType
    from .typing import HandlerGuard
    from .typing import HandlerMiddleware
    from .typing import Handlers
    from .typing import MiddlewareCallable
    from .typing import Scope

_router_count = count(1).__next__


class Router(BaseHandler[UpdateT]):
    """Base class for routers."""

    __slots__ = (
        "_name",
        "_frozen",
        "_routers",
        "_handlers",
        "_handler_guards",
        "_handler_middleware",
        "_middleware_stack",
        "_sort_handlers",
        "_sort_routers",
        "_exclude_locals",
        "_stop_propagate",
    )

    def __init__(  # yapf: disable
        self,
        *,
        filters: Sequence[BaseFilter[UpdateT]] | None = None,
        handlers: Sequence[Handler[UpdateT]] | None = None,
        dependencies: dict[str, Provide] | None = None,
        handler_guards: Sequence[HandlerGuard[UpdateT] | tuple[HandlerGuard[UpdateT], Scope]] | None = None,  # noqa: E501
        handler_middleware: Sequence[HandlerMiddleware[UpdateT] | tuple[HandlerMiddleware[UpdateT], Scope]] | None = None,  # noqa: E501
        routers: Sequence[Router[UpdateT]] | None = None,
        guards: Sequence[HandlerGuard[UpdateT]] | None = None,
        middleware: Sequence[HandlerMiddleware[UpdateT]] | None = None,
        name: str | None = None,
        sort_handlers: bool = True,
        sort_routers: bool = True,
        stop_propagate: bool | None = None,
    ) -> None:
        """
        :param handlers: *Optional.* A sequence of :class:`~yatbaf.typing.Handler`.
        :param dependencies: *Optional.* A mapping of dependency providers.
        :param handler_guards: *Optional.* A sequence of :class:`~yatbaf.typing.HandlerGuard`.
        :param handler_middleware: *Optional.* A sequence of :class:`~yatbaf.typing.HandlerMiddleware`.
        :param routers: *Optional.* A sequence of :class:`Router`.
        :param guards: *Optional.* A sequence of :class:`~yatbaf.typing.HandlerGuard`.
        :param middleware: *Optional.* A sequence of :class:`~yatbaf.typing.HandlerMiddleware`.
        :param name: *Optional.* Router name.
        :param sort_handlers: *Optional.* Pass ``False``, if you don't want to
            sort handlers.
        :param sort_routers: *Optional.* Pass ``False``, if you don't want to
            sort routers.
        :param stop_propagate: *Optional.* Pass ``True`` to stop propagate to
            next router even if no handler is found. Default ``bool(filters)``.
        """  # noqa: E501
        super().__init__(
            filters=filters,
            guards=guards,
            middleware=middleware,
            dependencies=dependencies,
        )

        self._name: Final[str] = name if name else f"router-{_router_count()}"
        self._frozen = False
        self._exclude_locals = False
        self._middleware_stack: HandlerCallableType[UpdateT] = self._handle
        self._sort_handlers = sort_handlers
        self._sort_routers = sort_routers
        self._stop_propagate = (
            stop_propagate if stop_propagate is not None else bool(filters)
        )

        self._handler_guards = list(handler_guards or [])
        self._handler_middleware = list(handler_middleware or [])

        self._handlers: list[Handler[UpdateT]] = []
        for handler in (handlers or []):
            self.add_handler(handler)

        self._routers: list[Router[UpdateT]] = []
        for router in (routers or []):
            self.add_router(router)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}[name={self._name}]>"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Router) and (  # yapf: disable
            other is self or (
                other.update_type == self.update_type
                and other._handlers == self._handlers
                and other._handler_guards == self._handler_guards
                and other._handler_middleware == self._handler_middleware
                and other._routers == self._routers
                and other._guards == self._guards
                and other._middleware == self._middleware
                and other._dependencies == self._dependencies
            )
        )

    @property
    def name(self) -> str:
        """Router name."""
        return self._name

    def add_guard(
        self, obj: HandlerGuard[UpdateT], /, scope: Scope = "handler"
    ) -> None:
        """Add a new guard.

        :param obj: :class:`~yatbaf.typing.HandlerGuard` object.
        :param scope: *Optional.* Scope of guard.
        :raises FrozenInstanceError: If you try to register a Guard after Bot
            object has been initialized.
        """
        if self._frozen:
            raise FrozenInstanceError(
                "It is not possible to add a new Guard at runtime "
                "after Bot object has been initialized."
            )

        if scope == "router":
            self._guards.append(obj)
        else:
            self._handler_guards.append(
                obj if scope != "local" else (obj, "local")
            )

    def add_middleware(  # yapf: disable
        self, obj: HandlerMiddleware[UpdateT], /, scope: Scope = "handler"
    ) -> None:
        """Add a new middleware.

        Usage::

            def middleware(handler: HandlerCallableType[UpdateT]) -> HandlerCallableType[UpdateT]:
                async def wrapper(update: UpdateT) -> None:
                    await handler(update)
                return wrapper

            router.add_middleware(middleware)

        :param obj: :class:`~yatbaf.typing.HandlerMiddleware` object.
        :param scope: *Optional.* Scope of middleware.
        :raises FrozenInstanceError: If you try to register a Middleware after
            Bot object has been initialized.
        """  # noqa: E501
        if self._frozen:
            raise FrozenInstanceError(
                "It is not possible to add a new Middleware at runtime "
                "after Bot object has been initialized."
            )

        if scope == "router":
            self._middleware.append(obj)
        else:
            self._handler_middleware.append(
                obj if scope != "local" else (obj, "local")
            )

    @overload
    def add_handler(self, handler: Handler[UpdateT], /) -> None:
        ...

    @overload
    def add_handler(
        self,
        handler: HandlerCallable[UpdateT],
        *,
        filters: Sequence[BaseFilter[UpdateT]] | None = None,
        middleware: Sequence[HandlerMiddleware[UpdateT]] | None = None,
        guards: Sequence[HandlerGuard[UpdateT]] | None = None,
        dependencies: dict[str, Provide] | None = None,
    ) -> None:
        ...

    def add_handler(
        self,
        handler: HandlerCallable[UpdateT] | Handler[UpdateT],
        *,
        filters: Sequence[BaseFilter[UpdateT]] | None = None,
        middleware: Sequence[HandlerMiddleware[UpdateT]] | None = None,
        guards: Sequence[HandlerGuard[UpdateT]] | None = None,
        dependencies: dict[str, Provide] | None = None,
    ) -> None:
        """Use this method to register a new handler.

        :param handler: :class:`~yatbaf.handler.Handler` or :class:`~yatbaf.typing.HandlerCallable`.
        :param filters: *Optional.* A sequence of :class:`~yatbaf.filters.base.BaseFilter`.
        :param middleware: *Optional.* A sequence of :class:`~yatbaf.typing.HandlerMiddleware`.
        :param guards: *Optional.* A sequence of :class:`~yatbaf.typing.HandlerGuard`.
        :param dependencies: *Optional.* A mapping of dependency providers.
        :raises FrozenInstanceError: If you try to register a Handler after Bot
            object has been initialized.
        """  # noqa: E501
        if self._frozen:
            raise FrozenInstanceError(
                f"{self!r} is frozen. It is not possible to add a new Handler "
                "at runtime after Bot object has been initialized."
            )

        if not isinstance(handler, Handler):
            handler = Handler(
                fn=handler,
                update_type=self.update_type,
                filters=filters,
                middleware=middleware,
                guards=guards,
                dependencies=dependencies,
            )

        self._validate_handler(handler)
        if handler.parent or (handler in self._handlers):
            warn_duplicate(handler, self)
            return

        self._handlers.append(handler)
        handler.parent = self

    def add_router(self, router: Router[UpdateT], /) -> None:
        """Use this method to add a nested router.

        .. warning::

            The nested router must handle the same type of updates as the
            current one.

        :param router: Instance of :class:`Router`
        :raises ValueError: If ``router`` already registered in another router
            or router type is different.
        :raises FrozenInstanceError: If you try to register a Router after Bot
            object has been initialized.
        """
        if self._frozen:
            raise FrozenInstanceError(
                f"{self!r} is frozen. It is not possible to add a new Router "
                "at runtime after Bot object has been initialized."
            )

        self._validate_router(router)
        if router.parent or (router in self._routers):
            warn_duplicate(router, self)
            return

        self._routers.append(router)
        router.parent = self

    def _validate_router(self, router: Router[UpdateT]) -> None:
        if router is self:
            raise ValueError(f"It is not possible to add {router!r} to itself.")
        if self.update_type != router.update_type:
            raise ValueError(
                f"Incompatible type! Cannot add {router!r} to {self!r}"
            )
        if (parent := router.parent) and parent is not self:
            raise ValueError(f"{router!r} already registered in {parent!r}")

    def _validate_handler(self, handler: Handler[UpdateT]) -> None:
        if handler.update_type != self.update_type:
            raise ValueError(
                f"Incompatible type! Cannot add {handler!r} to {self!r}"
            )
        if (parent := handler.parent) and parent is not self:
            raise ValueError(f"{handler!r} alredy registered in {parent!r}")

    @overload
    def __call__(  # yapf: disable
        self, __fn: HandlerCallable[UpdateT], /
    ) -> HandlerCallable[UpdateT]:
        ...

    @overload
    def __call__(
        self,
        *,
        filters: Sequence[BaseFilter[UpdateT]] | None = None,
        middleware: Sequence[HandlerMiddleware[UpdateT]] | None = None,
        guards: Sequence[HandlerGuard[UpdateT]] | None = None,
        dependencies: dict[str, Provide] | None = None,
    ) -> MiddlewareCallable[HandlerCallable[UpdateT]]:
        ...

    def __call__(  # yapf: disable
        self,
        __fn: HandlerCallable[UpdateT] | None = None,
        *,
        filters: Sequence[BaseFilter[UpdateT]] | None = None,
        middleware: Sequence[HandlerMiddleware[UpdateT]] | None = None,
        guards: Sequence[HandlerGuard[UpdateT]] | None = None,
        dependencies: dict[str, Provide] | None = None,
    ) -> MiddlewareCallable[HandlerCallable[UpdateT]] | HandlerCallable[UpdateT]:  # noqa: E501
        """Handler decorator.

        See :meth:`add_handler`.

        Use this decorator to register a new handler::

            @router
            async def handler(message):
                # handle any message
                ...


            @router(filters=[Command("foo")])
            async def handler(message):
                # handle command `/foo`
                ...
        """

        def wrapper(fn: HandlerCallable[UpdateT]) -> HandlerCallable[UpdateT]:
            self.add_handler(
                handler=fn,
                filters=filters,
                middleware=middleware,
                guards=guards,
                dependencies=dependencies,
            )
            return fn

        if __fn is not None:
            return wrapper(__fn)
        return wrapper

    # yapf: disable
    @overload
    def middleware(
        self, __fn: HandlerMiddleware[UpdateT], /
    ) -> HandlerMiddleware[UpdateT]:
        ...

    @overload
    def middleware(
        self, __scope: Scope = ..., /, *args: Any, **kwargs: Any
    ) -> MiddlewareCallable[HandlerMiddleware[UpdateT]]:
        ...

    def middleware(
        self,
        __fn_scope: HandlerMiddleware[UpdateT] | Scope = "handler", /,
        *args: Any,
        **kwargs: Any,
    ) -> MiddlewareCallable[HandlerMiddleware[UpdateT]] | HandlerMiddleware[UpdateT]:  # noqa: E501
        """Middleware decorator.

        Use this decorator to register a new middleware::

            @router.middleware
            def middleware(handler: HandlerCallableType[Message]) -> HandlerCallableType[Message]:
                async def wrapper(message: Message) -> None:
                    await handler(message)
                return wrapper


            @router.middleware("handler", 1, y=2)
            def middleware(handler: HandlerCallableType[Message], x, y) -> HandlerCallableType[Message]:
                async def wrapper(message: Message) -> None:
                    log.info(f"{x=}, {y=}")
                    await handler(message)
                return wrapper
        """  # noqa: E501
        # yapf: enable

        def wrapper(
            fn: HandlerMiddleware[UpdateT]
        ) -> HandlerMiddleware[UpdateT]:
            scope: Scope = __fn_scope  # type: ignore[assignment]
            if args or kwargs:
                self.add_middleware(Middleware(fn, *args, **kwargs), scope)
            else:
                self.add_middleware(fn, scope)
            return fn

        if not isinstance(__fn_scope, str):
            fn = __fn_scope
            __fn_scope = "handler"
            return wrapper(fn)
        return wrapper

    # yapf: disable
    @overload
    def guard(self, __fn: HandlerGuard[UpdateT], /) -> HandlerGuard[UpdateT]:
        ...

    @overload
    def guard(
        self, __scope: Scope = ..., /
    ) -> Callable[[HandlerGuard[UpdateT]], HandlerGuard[UpdateT]]:
        ...

    def guard(
        self, __fn_scope: HandlerGuard[UpdateT] | Scope = "handler"
    ) -> HandlerGuard[UpdateT] | Callable[[HandlerGuard[UpdateT]], HandlerGuard[UpdateT]]:  # noqa: E501
        """Guard decorator.

        Use this decorator to register a guard for handlers or router::

            users = [...]

            @router.guard
            async def guard(message: Message) -> None:
                if message.from_.id not in users:
                    raise GuardException


            @router.guard("router")
            async def guard(message: Message) -> None:
                if message.from_.id not in users:
                    raise GuardException
        """
        # yapf: enable

        def wrapper(fn: HandlerGuard[UpdateT]) -> HandlerGuard[UpdateT]:
            scope: Scope = __fn_scope  # type: ignore[assignment]
            self.add_guard(fn, scope)
            return fn

        if not isinstance(__fn_scope, str):
            fn = __fn_scope
            __fn_scope = "handler"
            return wrapper(fn)
        return wrapper

    async def _find_handler(self, update: UpdateT) -> Handler[UpdateT] | None:
        for handler in self._handlers:
            if await handler.match(update):
                return handler
        return None

    async def _handle(self, update: UpdateT, /) -> None:
        try:
            if self._guards:
                await self._check_guards(update)
        except GuardException:
            update.__usrctx__["handled"] = self._stop_propagate
            return

        if handler := await self._find_handler(update):
            update.__usrctx__["handled"] = True
            await handler.handle(update)
            return

        for router in self._routers:
            await router.handle(update)
            if update.__usrctx__["handled"]:
                return

        # skip other routers
        update.__usrctx__["handled"] = self._stop_propagate

    async def handle(self, update: UpdateT, /) -> None:
        update.__usrctx__["handled"] = False
        if not self._filters or await self.match(update):
            await self._middleware_stack(update)

    def _get_guards(self) -> Iterable[HandlerGuard[UpdateT]]:
        return _unpack(self._handler_guards, self._exclude_locals)

    def _get_middleware(self) -> Iterable[HandlerMiddleware[UpdateT]]:
        return _unpack(self._handler_middleware, self._exclude_locals)

    def _prepare_handlers(self) -> None:
        for handler in self._handlers:
            handler.on_registration()

        if self._sort_handlers:
            self._handlers.sort(reverse=True)
        self._exclude_locals = True

    def _prepare_routers(self) -> None:
        for router in self._routers:
            router.on_registration()

        if self._sort_routers:
            self._routers.sort(reverse=True)

    def on_registration(self) -> None:
        self._resolve_filters()
        self._prepare_handlers()
        self._resolve_guards()
        self._resolve_middleware()
        self._prepare_routers()

        self._middleware_stack = wrap_middleware(
            self._handle,
            self._middleware,  # type: ignore[arg-type]
        )
        self._frozen = True


T = TypeVar("T")


def _unpack(
    objs: Sequence[T | tuple[T, str]],
    exclude_local: bool = False,
) -> Iterator[T]:
    for obj in objs:
        scope = ""
        if isinstance(obj, tuple):
            obj, scope = obj
            if scope == "local" and exclude_local:
                continue
        yield cast("T", obj)


@final
class OnMessage(Router[Message]):
    """message router.

    See :attr:`Update.message <yatbaf.types.update.Update.message>`
    """

    __slots__ = ()
    update_type: UpdateType = UpdateType.MESSAGE


@final
class OnEditedMessage(Router[Message]):
    """edited_message router.

    See :attr:`Update.edited_message <yatbaf.types.update.Update.edited_message>`
    """  # noqa: E501

    __slots__ = ()
    update_type: UpdateType = UpdateType.EDITED_MESSAGE


@final
class OnChannelPost(Router[Message]):
    """channel_post router.

    See :attr:`Update.channel_post <yatbaf.types.update.Update.channel_post>`
    """

    __slots__ = ()
    update_type: UpdateType = UpdateType.CHANNEL_POST


@final
class OnEditedChannelPost(Router[Message]):
    """edited_channel_post router.

    See :attr:`Update.edited_channel_post <yatbaf.types.update.Update.edited_channel_post>`
    """  # noqa: E501

    __slots__ = ()
    update_type: UpdateType = UpdateType.EDITED_CHANNEL_POST


@final
class OnBusinessConnection(Router[BusinessConnection]):
    """business_connection router.

    See :attr:`~yatbaf.types.update.Update.business_connection`
    """  # noqa: E501

    __slots__ = ()
    update_type: UpdateType = UpdateType.BUSINESS_CONNECTION


@final
class OnBusinessMessage(Router[Message]):
    """business_message router.

    See :attr:`~yatbaf.types.update.Update.business_message`
    """  # noqa: E501

    __slots__ = ()
    update_type: UpdateType = UpdateType.BUSINESS_MESSAGE


@final
class OnEditedBusinessMessage(Router[Message]):
    """edited_business_message router.

    See :attr:`~yatbaf.types.update.Update.edited_business_message`
    """  # noqa: E501

    __slots__ = ()
    update_type: UpdateType = UpdateType.EDITED_BUSINESS_MESSAGE


@final
class OnDeletedBusinessMessages(Router[BusinessMessagesDeleted]):
    """deleted_business_messages router.

    See :attr:`~yatbaf.types.update.Update.deleted_business_messages`
    """  # noqa: E501

    __slots__ = ()
    update_type: UpdateType = UpdateType.DELETED_BUSINESS_MESSAGES


@final
class OnMessageReaction(Router[MessageReactionUpdated]):
    """message_reaction router.

    See :attr:`Update.message_reaction <yatbaf.types.update.Update.message_reaction>`
    """  # noqa: E501

    __slots__ = ()
    update_type: UpdateType = UpdateType.MESSAGE_REACTION


@final
class OnMessageReactionCount(Router[MessageReactionCountUpdated]):
    """message_reaction router.

    See :attr:`Update.message_reaction_count <yatbaf.types.update.Update.message_reaction_count>`
    """  # noqa: E501

    __slots__ = ()
    update_type: UpdateType = UpdateType.MESSAGE_REACTION_COUNT


@final
class OnInlineQuery(Router[InlineQuery]):
    """inline_query router.

    See :attr:`Update.inline_query <yatbaf.types.update.Update.inline_query>`
    """

    __slots__ = ()
    update_type: UpdateType = UpdateType.INLINE_QUERY


@final
class OnChosenInlineResult(Router[ChosenInlineResult]):
    """chosen_inline_result router.

    See :attr:`Update.chosen_inline_result <yatbaf.types.update.Update.chosen_inline_result>`
    """  # noqa: E501

    __slots__ = ()
    update_type: UpdateType = UpdateType.CHOSEN_INLINE_RESULT


@final
class OnCallbackQuery(Router[CallbackQuery]):
    """callback_query router.

    See :attr:`Update.callback_query <yatbaf.types.update.Update.callback_query>`
    """  # noqa: E501

    __slots__ = ()
    update_type: UpdateType = UpdateType.CALLBACK_QUERY


@final
class OnShippingQuery(Router[ShippingQuery]):
    """shipping_query router.

    See :attr:`Update.shipping_query <yatbaf.types.update.Update.shipping_query>`
    """  # noqa: E501

    __slots__ = ()
    update_type: UpdateType = UpdateType.SHIPPING_QUERY


@final
class OnPreCheckoutQuery(Router[PreCheckoutQuery]):
    """pre_checkout_query router.

    See :attr:`Update.pre_checkout_query <yatbaf.types.update.Update.pre_checkout_query>`
    """  # noqa: E501

    __slots__ = ()
    update_type: UpdateType = UpdateType.PRE_CHECKOUT_QUERY


@final
class OnPoll(Router[Poll]):
    """poll router.

    See :attr:`Update.poll <yatbaf.types.update.Update.poll>`
    """

    __slots__ = ()
    update_type: UpdateType = UpdateType.POLL


@final
class OnPollAnswer(Router[PollAnswer]):
    """poll_answer router.

    See :attr:`Update.poll_answer <yatbaf.types.update.Update.poll_answer>`
    """

    __slots__ = ()
    update_type: UpdateType = UpdateType.POLL_ANSWER


@final
class OnMyChatMember(Router[ChatMemberUpdated]):
    """my_chat_member router.

    See :attr:`Update.my_chat_member <yatbaf.types.update.Update.my_chat_member>`
    """  # noqa: E501

    __slots__ = ()
    update_type: UpdateType = UpdateType.MY_CHAT_MEMBER


@final
class OnChatMember(Router[ChatMemberUpdated]):
    """chat_member router.

    See :attr:`Update.chat_member <yatbaf.types.update.Update.chat_member>`
    """

    __slots__ = ()
    update_type: UpdateType = UpdateType.CHAT_MEMBER


@final
class OnChatJoinRequest(Router[ChatJoinRequest]):
    """chat_join_request router.

    See :attr:`Update.chat_join_request <yatbaf.types.update.Update.chat_join_request>`
    """  # noqa: E501

    __slots__ = ()
    update_type: UpdateType = UpdateType.CHAT_JOIN_REQUEST


@final
class OnChatBoost(Router[ChatBoostUpdated]):
    """chat_boost router.

    See :attr:`Update.chat_boost <yatbaf.types.update.Update.chat_boost>`
    """  # noqa: E501

    __slots__ = ()
    update_type: UpdateType = UpdateType.CHAT_BOOST


@final
class OnRemovedChatBoost(Router[ChatBoostRemoved]):
    """removed_chat_boost router.

    See :attr:`Update.removed_chat_boost <yatbaf.types.update.Update.removed_chat_boost>`
    """  # noqa: E501

    __slots__ = ()
    update_type: UpdateType = UpdateType.REMOVED_CHAT_BOOST


_router_map: dict[str, type[Router]] = {
    "message": OnMessage,
    "edited_message": OnEditedMessage,
    "channel_post": OnChannelPost,
    "edited_channel_post": OnEditedChannelPost,
    "business_connection": OnBusinessConnection,
    "business_message": OnBusinessMessage,
    "edited_business_message": OnEditedBusinessMessage,
    "deleted_business_messages": OnDeletedBusinessMessages,
    "message_reaction": OnMessageReaction,
    "message_reaction_count": OnMessageReactionCount,
    "inline_query": OnInlineQuery,
    "chosen_inline_result": OnChosenInlineResult,
    "callback_query": OnCallbackQuery,
    "shipping_query": OnShippingQuery,
    "pre_checkout_query": OnPreCheckoutQuery,
    "poll": OnPoll,
    "poll_answer": OnPollAnswer,
    "my_chat_member": OnMyChatMember,
    "chat_member": OnChatMember,
    "chat_join_request": OnChatJoinRequest,
    "chat_boost": OnChatBoost,
    "removed_chat_boost": OnRemovedChatBoost,
}


def parse_handlers(
    *,
    handlers: Sequence[Handler] | None = None,
    routers: Sequence[Router] | None = None,
    dependencies: dict[str, Provide] | None = None,
) -> Handlers:
    """:meta private:"""
    handlers = list(handlers or [])
    routers = list(routers or [])
    dependencies = dependencies or {}
    if not handlers and not routers:
        raise ValueError("You must use at least one handler or router.")

    # yapf: disable
    tmp: dict[str, tuple[list[Handler], list[Router]]] = (
        defaultdict(lambda: ([], []))
    )
    # yapf: enable
    for handler_ in handlers:
        tmp[handler_.update_type][0].append(handler_)

    for router_ in routers:
        tmp[router_.update_type][1].append(router_)

    result = {}
    for type_, (h, r) in tmp.items():
        router: BaseHandler
        if h:
            router = _router_map[type_](
                handlers=h,
                routers=r,
                dependencies=dependencies,
            )
        elif r:
            if len(r) == 1:
                router = r[0]
                router._dependencies = {
                    **dependencies,
                    **router._dependencies,
                }
            else:
                router = _router_map[type_](
                    routers=r,
                    dependencies=dependencies,
                )
        result[type_] = router
        router.on_registration()

    return cast("Handlers", result)



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .business_opening_hours_interval import BusinessOpeningHoursInterval


@final
class BusinessOpeningHours(TelegramType):
    """See: https://core.telegram.org/bots/api#businessopeninghours"""

    time_zone_name: str
    """Unique name of the time zone for which the opening hours are defined."""

    opening_hours: list[BusinessOpeningHoursInterval]
    """List of time intervals describing business opening hours."""



from __future__ import annotations

__all__ = (
    "ResponseError",
    "ResponseOk",
    "Request",
    "HTTPResponse",
)

from typing import Any
from typing import Generic
from typing import TypeVar
from typing import final

from msgspec import Struct

from .types import ResponseParameters
from .typing import NoneStr

T = TypeVar("T")


@final
class Request(Struct):
    """Request object"""

    url: str
    """URL."""

    method: str
    """Telegram method name."""

    data: dict[str, Any] | None = None
    """Request POST data."""

    files: dict[str, Any] | None = None
    """Request files."""

    content: bytes | None = None
    """Json (bytes) content"""

    headers: dict[str, str] | None = None
    """Request headers."""


@final
class HTTPResponse(Struct, Generic[T]):
    """This object represents HTTP response."""

    status_code: int
    """Request status code."""

    content: T
    """Response content."""


class ApiResponse(Struct):
    """Base object for api response.

    See: https://core.telegram.org/bots/api#making-requests
    """

    ok: bool
    """Request status."""


@final
class ResponseOk(ApiResponse, Generic[T]):
    """This object represents successful api response."""

    result: T
    """Result of the query."""

    description: NoneStr = None
    """Human-readable description of the result."""


@final
class ResponseError(ApiResponse):
    """This object represents unsuccessful api response."""

    error_code: int
    """HTTP status code."""

    description: str
    """Human-readable description of the error."""

    parameters: ResponseParameters | None = None
    """See :class:`yatbaf.types.ResponseParameters`."""



# Based on litestar's di, distributed under MIT license.
# github.com/litestar-org/litestar/blob/main/LICENSE

from __future__ import annotations

__all__ = ("Provide",)

import asyncio
import inspect
from typing import TYPE_CHECKING
from typing import Any
from typing import Final

from .exceptions import DependencyError

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator
    from collections.abc import Callable

    from .typing import HandlerDependency

RESERVED_KWARGS: Final = frozenset(["update"])


class Provide:
    """Wrapper class for dependency injection."""

    __slots__ = (
        "dependency",
        "dependency_kwargs",
        "sync_call",
        "is_generator",
    )

    def __init__(self, dependency: HandlerDependency) -> None:
        """
        :param dependency: Callable to call or class to instantiate. The result
            is then injected as a dependency.
        """
        if not callable(dependency):
            raise DependencyError("dependency must be callable")

        self.dependency = dependency
        self.dependency_kwargs = frozenset(get_parameters(dependency))

        is_class = inspect.isclass(dependency)
        is_generator = not is_class and _is_async_generator(dependency)

        if not (is_class or is_generator):
            if not _is_async_callable(dependency):
                raise DependencyError("dependency must be async callable")

        self.sync_call = is_class or is_generator
        self.is_generator = is_generator

    async def __call__(self, **kwargs: Any) -> Any:
        if self.sync_call:
            return self.dependency(**kwargs)
        return await self.dependency(**kwargs)  # type: ignore[misc]

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Provide) and (  # yapf: disable
            other is self or other.dependency == self.dependency
        )


class Dependency:

    __slots__ = (
        "key",
        "provider",
        "dependencies",
    )

    def __init__(
        self, key: str, provider: Provide, dependencies: list[Dependency]
    ) -> None:
        self.key = key
        self.provider = provider
        self.dependencies = dependencies

    def __eq__(self, other: object) -> bool:
        return isinstance(other, Dependency) and (
            other is self or (  # yapf: disable
                other.key == self.key
                and other.provider == self.provider
                and other.dependencies == self.dependencies
            )
        )

    def __hash__(self) -> int:
        return hash(self.key)


class CleanupGroup:

    __slots__ = ("generators",)

    def __init__(self) -> None:
        self.generators: list[AsyncGenerator[Any, None]] = []

    def add(self, gen: AsyncGenerator[Any, None]) -> None:
        self.generators.append(gen)

    async def cleanup(self) -> None:
        if not self.generators:
            return

        if len(self.generators) == 1:
            await anext(self.generators[0], None)
            return

        async with asyncio.TaskGroup() as tg:
            for gen in self.generators:
                tg.create_task(anext(gen, None))

    async def throw(self, exc: BaseException) -> None:
        for gen in self.generators:
            try:
                await gen.athrow(exc)
            except StopAsyncIteration:
                continue


def get_parameters(fn: Callable[..., Any]) -> list[str]:
    return list(inspect.signature(fn).parameters)


def is_reserved_key(key: str) -> bool:
    return key in RESERVED_KWARGS


def _is_async_generator(fn: Callable[..., Any]) -> bool:
    return (
        inspect.isasyncgenfunction(fn)
        or inspect.isasyncgenfunction(fn.__call__)  # type: ignore[operator]
    )


def _is_async_callable(fn: Callable[..., Any]) -> bool:
    return (
        inspect.iscoroutinefunction(fn)
        or inspect.iscoroutinefunction(fn.__call__)  # type: ignore[operator]
    )


def validate_provider(
    key: str,
    provider: Provide,
    dependencies: dict[str, Provide],
) -> None:
    if is_reserved_key(key):
        raise DependencyError(f"name {key!r} is reserved.")

    for k, v in dependencies.items():
        if v == provider and k != key:
            raise DependencyError(
                f"Provider {key!r} is already registered as {k!r}! "
                "Use the same name to override provider."
            )


async def _resolve_dependency(
    dependency: Dependency, kwargs: dict[str, Any], cleanup_group: CleanupGroup
) -> None:
    dependency_kwargs = {
        k: kwargs[k]
        for k in dependency.provider.dependency_kwargs
    }
    value = await dependency.provider(**dependency_kwargs)
    if dependency.provider.is_generator:
        cleanup_group.add(value)
        value = await anext(value)
    kwargs[dependency.key] = value


async def resolve_dependencies(
    dependencies: list[set[Dependency]], kwargs: dict[str, Any]
) -> CleanupGroup:
    cleanup_group = CleanupGroup()
    for batch in dependencies:
        if len(batch) == 1:
            await _resolve_dependency(next(iter(batch)), kwargs, cleanup_group)
        else:
            try:
                async with asyncio.TaskGroup() as tg:
                    for dependency in batch:
                        tg.create_task(
                            _resolve_dependency(
                                dependency,
                                kwargs,
                                cleanup_group,
                            )
                        )
            except ExceptionGroup as eg:
                raise eg.exceptions[0] from eg

    return cleanup_group


def create_dependency_graph(
    key: str, dependencies: dict[str, Provide]
) -> Dependency:
    try:
        provider = dependencies[key]
    except KeyError:
        raise DependencyError(f"Provider is missing: {key!r}") from None

    sub_dependencies = [
        k for k in provider.dependency_kwargs if k not in RESERVED_KWARGS
    ]
    return Dependency(
        key,
        provider,
        dependencies=[
            create_dependency_graph(k, dependencies) for k in sub_dependencies
        ]
    )


def create_dependency_batches(
    expected_dependencies: set[Dependency]
) -> list[set[Dependency]]:
    dependencies_to: dict[Dependency, set[Dependency]] = {}
    for dependency in expected_dependencies:
        if dependency not in dependencies_to:
            map_dependencies_recursively(dependency, dependencies_to)

    batches = []
    while dependencies_to:
        current_batch = {  # yapf: disable
            dependency
            for dependency, remaining_sub_dependencies in dependencies_to.items()  # noqa: E501
            if not remaining_sub_dependencies
        }

        for dependency in current_batch:
            del dependencies_to[dependency]
            for others_dependencies in dependencies_to.values():
                others_dependencies.discard(dependency)

        batches.append(current_batch)

    return batches


def map_dependencies_recursively(
    dependency: Dependency,
    dependencies_to: dict[Dependency, set[Dependency]],
) -> None:
    dependencies_to[dependency] = set(dependency.dependencies)
    for sub in dependency.dependencies:
        if sub not in dependencies_to:
            map_dependencies_recursively(sub, dependencies_to)



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneStr

from .abc import TelegramType
from .photo_size import PhotoSize


@final
class SharedUser(TelegramType):
    """This object contains information about a user that was shared with the
    bot using a :class:`~.keyboard_button.KeyboardButtonRequestUser` button.

    See: https://core.telegram.org/bots/api#shareduser
    """

    user_id: int
    """Identifier of the shared user.

    .. note::

        The bot may not have access to the user and could be unable to use this
        identifier, unless the user is already known to the bot by some other
        means.
    """

    first_name: NoneStr = None
    """*Optional.* First name of the user, if the name was requested by the bot."""  # noqa: E501

    last_name: NoneStr = None
    """*Optional.* Last name of the user, if the name was requested by the bot."""  # noqa: E501

    username: NoneStr = None
    """*Optional.* Username of the user, if the username was requested by the
    bot.
    """

    photo: list[PhotoSize] | None = None
    """*Optional.* Available sizes of the chat photo, if the photo was requested
    by the bot.
    """



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import final

from yatbaf.types import Message

from .abc import TelegramMethod

if TYPE_CHECKING:
    from yatbaf.enums import ParseMode
    from yatbaf.types import LinkPreviewOptions
    from yatbaf.types import MessageEntity
    from yatbaf.types import ReplyParameters
    from yatbaf.typing import NoneBool
    from yatbaf.typing import NoneInt
    from yatbaf.typing import NoneStr
    from yatbaf.typing import ReplyMarkup


@final
class SendMessage(TelegramMethod[Message]):
    """See :meth:`yatbaf.bot.Bot.send_message`"""

    chat_id: str | int
    text: str
    business_connection_id: NoneStr = None
    message_thread_id: NoneInt = None
    parse_mode: ParseMode | None = None
    entities: list[MessageEntity] | None = None
    link_preview_options: LinkPreviewOptions | None = None
    disable_notification: NoneBool = None
    protect_content: NoneBool = None
    reply_parameters: ReplyParameters | None = None
    reply_markup: ReplyMarkup | None = None



from __future__ import annotations

from typing import TYPE_CHECKING

from .base import BaseFilter

if TYPE_CHECKING:
    from yatbaf.types import CallbackQuery
    from yatbaf.typing import FilterPriority


class Data(BaseFilter):
    """CallbackQuery data filter."""

    __slots__ = ("data",)

    def __init__(self, data: str) -> None:
        self.data = data

    @property
    def priority(self) -> FilterPriority:
        return {"content": (1, 100)}

    async def check(self, q: CallbackQuery) -> bool:
        return (d := q.data) is not None and d == self.data



from yatbaf.dispatcher import Dispatcher


def test_dispatcher():
    handlers = {}
    dispatcher = Dispatcher(handlers)
    assert dispatcher._middleware_stack == dispatcher._resolve
    assert not dispatcher._guards
    assert dispatcher._handlers is handlers



from __future__ import annotations

__all__ = ("Dispatcher",)

import logging
from typing import TYPE_CHECKING

from .abc import AbstractRouter
from .exceptions import GuardException
from .utils import ensure_unique
from .utils import wrap_middleware

if TYPE_CHECKING:
    from collections.abc import Sequence

    from .types import Update
    from .typing import Handlers
    from .typing import RouterGurad
    from .typing import RouterMiddleware

log = logging.getLogger(__name__)


class Dispatcher(AbstractRouter):
    """Dispatcher object."""

    __slots__ = (
        "_handlers",
        "_guards",
        "_middleware_stack",
    )

    def __init__(
        self,
        handlers: Handlers,
        guards: Sequence[RouterGurad] | None = None,
        middleware: Sequence[RouterMiddleware] | None = None,
    ) -> None:
        """
        :param handlers: Handlers map.
        :param guards: *Optional.* A sequence of :class:`~yatbaf.typing.RouterGurad`.
        :param middleware: *Optional.* A sequence of :class:`~yatbaf.typing.RouterMiddleware`.
        """  # noqa: E501
        self._handlers: Handlers = handlers
        self._guards = ensure_unique(guards or [])
        self._middleware_stack = wrap_middleware(
            self._resolve,
            ensure_unique(middleware or []),  # type: ignore[arg-type]
        )

    async def _check_guards(self, update: Update) -> None:
        for fn in self._guards:
            await fn(update)

    async def _resolve(self, update: Update, /) -> None:
        if handler := self._handlers.get(update.event_type):
            try:
                if self._guards:
                    await self._check_guards(update)
            except GuardException:
                return

            await handler.handle(update.event)  # type: ignore[attr-defined]

    async def resolve(self, update: Update, /) -> None:
        await self._middleware_stack(update)



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .user import User


@final
class BusinessConnection(TelegramType):
    """Describes the connection of the bot with a business account.

    See: https://core.telegram.org/bots/api#businessconnection
    """

    id: str
    """Unique identifier of the business connection."""

    user: User
    """Business account user that created the business connection."""

    user_chat_id: int
    """Identifier of a private chat with the user who created the business
    connection.
    """

    date: int
    """Date the connection was established in Unix time."""

    can_reply: bool
    """``True``, if the bot can act on behalf of the business account in chats
    that were active in the last 24 hours.
    """

    is_enabled: bool
    """``True``, if the connection is active."""



from __future__ import annotations

from typing import final

from .abc import TelegramMethod


@final
class DeleteStickerFromSet(TelegramMethod[bool]):
    """See :meth:`yatbaf.bot.Bot.delete_sticker_from_set`"""

    sticker: str



from __future__ import annotations

from typing import final

from .abc import TelegramType


@final
class ChatBoostAdded(TelegramType):
    """This object represents a service message about a user boosting a chat.

    See: https://core.telegram.org/bots/api#chatboostadded
    """

    boost_count: int
    """Number of boosts added by the user."""



import pytest

from yatbaf.di import Provide
from yatbaf.exceptions import DependencyError
from yatbaf.handler import on_message
from yatbaf.router import OnMessage


async def data1():
    return 1


async def data2():
    return 2


def test_get_dependency_providers_empty():

    @on_message
    async def handler(_):
        pass

    assert handler._get_dependency_providers() == {}


def test_get_dependency_providers_local():

    @on_message(dependencies={"val": Provide(data1)})
    async def handler(_):
        pass

    assert handler._get_dependency_providers() == {"val": Provide(data1)}


def test_get_dependency_providers_router():

    @on_message
    async def handler(_):
        pass

    _ = OnMessage(
        dependencies={"val": Provide(data1)},
        handlers=[handler],
    )

    assert handler._get_dependency_providers() == {"val": Provide(data1)}


def test_get_dependency_providers_router_local():

    @on_message(dependencies={"d2": Provide(data2)})
    async def handler(_):
        pass

    _ = OnMessage(
        dependencies={"d1": Provide(data1)},
        handlers=[handler],
    )

    assert handler._get_dependency_providers() == {
        "d1": Provide(data1),
        "d2": Provide(data2),
    }


def test_get_dependency_providers_override():

    @on_message(dependencies={"val": Provide(data2)})
    async def handler(_):
        pass

    _ = OnMessage(
        dependencies={"val": Provide(data1)},
        handlers=[handler],
    )

    assert handler._get_dependency_providers() == {"val": Provide(data2)}


def test_get_dependency_providers_unique():

    @on_message(dependencies={"data": Provide(data1)})
    async def handler(_):
        pass

    _ = OnMessage(
        dependencies={"val": Provide(data1)},
        handlers=[handler],
    )

    with pytest.raises(DependencyError):
        handler._get_dependency_providers()


def test_resolve_dependencies_missing_provider():

    @on_message
    async def handler(_, val):  # noqa: U100
        pass

    with pytest.raises(DependencyError):
        handler._resolve_dependencies()


@pytest.mark.asyncio
async def test_provide_update_arg(message):

    async def provide_val():
        return "update-arg"

    @on_message(dependencies={"foo": Provide(provide_val)})
    async def handler(message, foo):
        message.ctx["test"] = foo

    handler.on_registration()
    await handler.handle(message)
    assert message.ctx["test"] == "update-arg"


@pytest.mark.asyncio
async def test_provide_update_kwarg(message):

    async def provide_val():
        return "foo"

    @on_message(dependencies={"foo": Provide(provide_val)})
    async def handler(message, foo, update):
        update.ctx["test"] = foo
        message.ctx["test"] += "bar"

    handler.on_registration()
    await handler.handle(message)
    assert message.ctx["test"] == "foobar"


@pytest.mark.asyncio
async def test_dependency_exception(message, mock_mark):

    async def provide_data():
        try:
            yield 1
        except ValueError:
            mock_mark()

    @on_message(dependencies={"data": Provide(provide_data)})
    async def handler(message, data):  # noqa: U100
        raise ValueError()

    handler.on_registration()
    with pytest.raises(ValueError):
        await handler.handle(message)
    mock_mark.assert_called_once()


@pytest.mark.asyncio
async def test_dependency_cleanup(message, mock_mark):

    async def provide_data():
        yield 1
        mock_mark()

    @on_message(dependencies={"data": Provide(provide_data)})
    async def handler(message, data):  # noqa: U100
        return

    handler.on_registration()
    await handler.handle(message)
    mock_mark.assert_called_once()


@pytest.mark.asyncio
async def test_handler_no_kwargs(message, mock_mark):

    async def provide_data():
        mock_mark()
        return 1

    @on_message(dependencies={"data": Provide(provide_data)})
    async def handler(message):  # noqa: U100
        pass

    handler.on_registration()
    await handler.handle(message)
    mock_mark.assert_not_called()



from __future__ import annotations

from typing import final

from yatbaf.types import BusinessConnection

from .abc import TelegramMethod


@final
class GetBusinessConnection(TelegramMethod[BusinessConnection]):
    """See: :meth:`~yatbaf.bot.Bot.get_business_connection`."""

    business_connection_id: str



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .shared_user import SharedUser


@final
class UsersShared(TelegramType):
    """This object contains information about the users whose identifiers were
    shared with the bot using a :class:`~yatbaf.types.keyboard_buton.KeyboardButtonRequestUsers` button.

    See: https://core.telegram.org/bots/api#usersshared
    """  # noqa: E501

    request_id: int
    """Identifier of the request."""

    users: list[SharedUser]
    """Information about users shared with the bot."""



from __future__ import annotations

from typing import final

from .abc import TelegramType
from .chat import Chat


@final
class BusinessMessagesDeleted(TelegramType):
    """This object is received when messages are deleted from a connected
    business account.

    See: https://core.telegram.org/bots/api#businessmessagesdeleted
    """

    business_connection_id: str
    """Unique identifier of the business connection."""

    chat: Chat
    """Information about a chat in the business account. The bot may not have
    access to the chat or the corresponding user.
    """

    message_ids: list[int]
    """List of identifiers of deleted messages in the chat of the business
    account.
    """



import pytest

from yatbaf.enums import UpdateType
from yatbaf.exceptions import BotWarning
from yatbaf.filters import Command
from yatbaf.handler import Handler
from yatbaf.router import OnMessage


def test_init_param(handler_fn):
    router = OnMessage(
        handlers=[
            handler := Handler(handler_fn, "message"),
        ],
    )
    assert router._handlers == [Handler(handler_fn, "message")]
    assert handler.parent is router


def test_add_handler_fn(handler_fn):
    router = OnMessage()
    router.add_handler(handler_fn)
    assert router._handlers == [Handler(handler_fn, UpdateType.MESSAGE)]


def test_add_handler_obj(handler_fn):
    router = OnMessage()
    router.add_handler(Handler(handler_fn, UpdateType.MESSAGE))
    assert router._handlers == [Handler(handler_fn, UpdateType.MESSAGE)]


def test_add_handler_type_error(handler_fn):
    router = OnMessage()
    handler = Handler(handler_fn, UpdateType.POLL)
    with pytest.raises(ValueError):
        router.add_handler(handler)


def test_add_handler_registered(handler_fn):
    handler = Handler(handler_fn, UpdateType.MESSAGE)
    _ = OnMessage(handlers=[handler])
    with pytest.raises(ValueError):
        OnMessage(handlers=[handler])


def test_add_handler_registered_same_router(handler_fn):
    router = OnMessage(
        handlers=[
            handler := Handler(handler_fn, UpdateType.MESSAGE),
        ]
    )
    with pytest.warns(BotWarning):
        router.add_handler(handler)
    assert router._handlers == [Handler(handler_fn, UpdateType.MESSAGE)]


def test_init_duplicate(handler_fn):
    with pytest.warns(BotWarning):
        router = OnMessage(
            handlers=[
                handler1 := Handler(handler_fn, UpdateType.MESSAGE),
                Handler(
                    handler_fn,
                    UpdateType.MESSAGE,
                    filters=[Command("foo")],
                ),
            ]
        )
    assert router._handlers == [Handler(handler_fn, UpdateType.MESSAGE)]
    assert router._handlers[0] is handler1


def test_add_handler_duplicate(handler_fn):
    router = OnMessage()
    handler1 = Handler(handler_fn, UpdateType.MESSAGE)
    handler2 = Handler(
        handler_fn,
        UpdateType.MESSAGE,
        filters=[Command("foo")],
    )
    router.add_handler(handler1)
    with pytest.warns(BotWarning):
        router.add_handler(handler2)
    assert router._handlers == [Handler(handler_fn, UpdateType.MESSAGE)]
    assert router._handlers[0] is handler1


def test_add_handler_duplicate_1(handler_fn):
    router = OnMessage()
    router.add_handler(handler_fn)
    with pytest.warns(BotWarning):
        router.add_handler(handler_fn, filters=[Command("foo")])
    assert router._handlers == [Handler(handler_fn, UpdateType.MESSAGE)]


def test_decorator_fallback():
    router = OnMessage()

    @router
    async def handler(_):  # noqa: U101
        pass

    assert router._handlers == [Handler(handler, UpdateType.MESSAGE)]


def test_decorator(guard_true):
    router = OnMessage()

    def middleware(h):

        async def w(u):
            await h(u)

        return w

    @router(
        middleware=[middleware],
        filters=[filter := Command("foo")],
        guards=[guard_true]
    )
    async def handler(_) -> None:  # noqa: U101
        pass

    assert router._handlers == [
        Handler(
            fn=handler,
            update_type=UpdateType.MESSAGE,
            middleware=[middleware],
        )
    ]
    assert router._handlers[0]._middleware == [middleware]
    assert router._handlers[0]._filters == (filter,)
    assert router._handlers[0]._guards == [guard_true]


def test_decorator_duplicate():
    router = OnMessage()

    with pytest.warns(BotWarning):

        @router
        @router(filters=[Command("foo")])
        async def handler(_):
            pass

    assert router._handlers == [Handler(handler, UpdateType.MESSAGE)]



import pytest

from yatbaf.handler import Handler
from yatbaf.router import OnMessage


def create_middleware(mark):

    def middleware(handler):

        async def wrapper(update):
            update.ctx["test"].append(mark)
            await handler(update)

        return wrapper

    return middleware


def create_callback(mark):

    async def handler(update):
        update.ctx["test"].append(mark)

    return handler


def test_resolve_middleware():
    mdlw1 = create_middleware(1)
    mdlw2 = create_middleware(2)

    router = OnMessage(
        middleware=[
            mdlw1,
            mdlw2,
            mdlw2,
        ],
    )
    router._resolve_middleware()
    assert router._middleware == [mdlw1, mdlw2]


@pytest.mark.asyncio
async def test_router_middleware_order(message):
    message.ctx["test"] = []
    router = OnMessage(
        handlers=[Handler(create_callback("h"), "message")],
        middleware=[create_middleware(f"rm{i}") for i in range(5)],
    )
    router.on_registration()
    await router.handle(message)
    assert message.ctx["test"] == [f"rm{i}" for i in range(5)] + ["h"]


@pytest.mark.asyncio
async def test_handler_middleware_order(message):
    message.ctx["test"] = []
    router = OnMessage(
        handlers=[Handler(create_callback("h"), update_type="message")],
        handler_middleware=[create_middleware(f"hm{i}") for i in range(5)],
    )
    router.on_registration()
    await router.handle(message)
    assert message.ctx["test"] == [f"hm{i}" for i in range(5)] + ["h"]


@pytest.mark.asyncio
async def test_router_handler_middleware_order(message):
    message.ctx["test"] = []
    router = OnMessage(
        handlers=[Handler(create_callback("h"), update_type="message")],
        handler_middleware=[create_middleware(f"hm{i}") for i in range(5)],
        middleware=[create_middleware(f"rm{i}") for i in range(5)],
    )
    router.on_registration()
    await router.handle(message)
    assert message.ctx["test"] == (  # yapf: disable
        [f"rm{i}" for i in range(5)] + [f"hm{i}" for i in range(5)] + ["h"]
    )



import pytest

from yatbaf.di import Provide
from yatbaf.exceptions import DependencyError


def sync_func():
    pass


async def func():
    return "func"


async def gen():
    yield "gen"


def sync_gen():
    yield "sync_gen"


async def func_params(fn_param):
    return fn_param


async def gen_params(gen_param):
    yield gen_param


class Class:
    val = "class-val"

    def __init__(self) -> None:
        self.val = "inst-val"

    @classmethod
    async def class_method(cls) -> int:
        return cls.val

    @staticmethod
    async def static_method() -> str:
        return "static-val"

    async def method(self) -> int:
        return self.val

    async def __call__(self):
        return "call-val"


class ClassParams:

    def __init__(self, init_param):
        self.init_param = init_param

    async def __call__(self, call_param):
        return call_param


@pytest.mark.parametrize("obj", [sync_gen, sync_func, "str", 123])
def test_sync_func(obj):
    with pytest.raises(DependencyError):
        Provide(obj)


@pytest.mark.parametrize(
    "dep,exp", [
        [func, False],
        [gen, True],
        [Class, True],
        [Class(), False],
    ]
)
def test_sync_call(dep, exp):
    assert Provide(dep).sync_call is exp


@pytest.mark.parametrize(
    "dep,exp", [
        [func, False],
        [gen, True],
        [Class, False],
        [Class(), False],
    ]
)
def test_is_generator(dep, exp):
    assert Provide(dep).is_generator is exp


@pytest.mark.parametrize(
    "dep,exp",
    [
        [func, frozenset()],
        [gen, frozenset()],
        [Class, frozenset()],
        [func_params, frozenset(["fn_param"])],
        [gen_params, frozenset(["gen_param"])],
        [ClassParams, frozenset(["init_param"])],
        [ClassParams(1), frozenset(["call_param"])],
    ]
)
def test_dependency_kwargs(dep, exp):
    assert Provide(dep).dependency_kwargs == exp


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "dep,exp",
    [
        [func, "func"],
        [Class(), "call-val"],
        [Class.class_method, "class-val"],
        [Class.static_method, "static-val"],
        [Class().method, "inst-val"],
    ]
)
async def test_result(dep, exp):
    assert await Provide(dep)() == exp


def test_eq():
    assert Provide(func) == Provide(func)
    assert Provide(func) != Provide(func_params)



import pytest

from yatbaf import filters as f
from yatbaf.router import OnMessage


@pytest.mark.asyncio
async def test_filters_true(message, user):
    user.id = 123
    router = OnMessage(filters=[f.User(123)])

    @router
    async def handler(message):
        message.ctx["test"] = True

    router.on_registration()
    await router.handle(message)
    assert message.ctx.get("test", False)


@pytest.mark.asyncio
async def test_filters_false(message, user):
    user.id = 321
    router = OnMessage(filters=[f.User(123)])

    @router
    async def handler(message):
        message.ctx["test"] = True

    router.on_registration()
    await router.handle(message)
    assert not message.ctx.get("test", False)



import pytest

from yatbaf.filters import Command


def test_empty():
    with pytest.raises(ValueError):
        Command()


def test_priority():
    assert Command("foo").priority == {"content": (1, 1000)}
    assert Command("foo", "bar").priority == {"content": (2, 1000)}


@pytest.mark.asyncio
async def test_true(message):
    message.text = "/ping"
    assert await Command("ping").check(message)


@pytest.mark.asyncio
async def test_false(message):
    message.text = "/pong"
    assert not await Command("ping").check(message)


@pytest.mark.asyncio
async def test_mix_true(message):
    message.text = "/pong"
    assert await Command("ping", "pong").check(message)


@pytest.mark.asyncio
async def test_text_is_none(message):
    message.text = None
    assert not await Command("ping").check(message)


@pytest.mark.asyncio
@pytest.mark.parametrize("filter", ["start", "/start", "START", "Start"])
async def test_case(message, filter):
    message.text = "/start"
    assert await Command(filter).check(message)



from __future__ import annotations

__all__ = (
    "AbstractClient",
    "AbstractRouter",
    "AbstractHandler",
)

from abc import ABC
from abc import abstractmethod
from typing import TYPE_CHECKING
from typing import Generic

from .typing import UpdateT

if TYPE_CHECKING:
    from collections.abc import AsyncIterator
    from contextlib import AbstractAsyncContextManager

    from .models import HTTPResponse
    from .models import Request
    from .types import Update


class AbstractClient(ABC):
    """Abstract http client."""

    __slots__ = ()

    @abstractmethod
    async def send_post(
        self,
        request: Request,  # noqa: U100
        *,
        timeout: float | None = None  # noqa: U100
    ) -> HTTPResponse[bytes]:
        """Send POST request.

        :param request: :class:`~yatbaf.models.Request` object.
        :param timeout: Request timeout.
        """
        pass

    @abstractmethod
    def file_stream(
        self,
        url: str,  # noqa: U100
        chunk_size: int,  # noqa: U100
    ) -> AbstractAsyncContextManager[HTTPResponse[AsyncIterator[bytes]]]:
        """Download file content.

        :param url: File URL.
        :param chunk_size: Chunk length in bytes.
        """
        pass

    @abstractmethod
    async def close(self) -> None:
        pass


class AbstractRouter(ABC):
    __slots__ = ()

    @abstractmethod
    async def resolve(self, update: Update, /) -> None:
        pass


class AbstractHandler(ABC, Generic[UpdateT]):
    __slots__ = ()

    @abstractmethod
    async def handle(self, update: UpdateT, /) -> None:
        pass



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneStr

from .abc import TelegramType
from .sticker import Sticker


@final
class BusinessIntro(TelegramType):
    """See: https://core.telegram.org/bots/api#businessintro"""

    title: NoneStr = None
    """*Optional.* Title text of the business intro."""

    message: NoneStr = None
    """*Optional.* Message text of the business intro."""

    sticker: Sticker | None = None
    """*Optional.* Sticker of the business intro."""



from __future__ import annotations

from typing import final

from yatbaf.typing import NoneStr

from .abc import TelegramType
from .photo_size import PhotoSize


@final
class ChatShared(TelegramType):
    """This object contains information about the chat whose identifier was
    shared with the bot using a
    :class:`yatbaf.types.KeyboardButtonRequestChat` button.

    See: https://core.telegram.org/bots/api#chatshared
    """

    request_id: int
    """Identifier of the request."""

    chat_id: int
    """Identifier of the shared chat.

    .. note::

        The bot may not have access to the chat and could be unable to use
        this identifier, unless the chat is already known to the bot by some
        other means.
    """

    title: NoneStr = None
    """*Optional.* Title of the chat, if the title was requested by the bot."""

    username: NoneStr = None
    """*Optional.* Username of the chat, if the username was requested by the
    bot and available.
    """

    photo: list[PhotoSize] | None = None
    """*Optional.* Available sizes of the chat photo, if the photo was requested
    by the bot.
    """



import pytest

from yatbaf.enums import UpdateType
from yatbaf.filters import Command
from yatbaf.filters import User
from yatbaf.handler import Handler


@pytest.mark.asyncio
async def test_match_empty_true(handler_fn, message):
    handler = Handler(
        handler_fn,
        update_type=UpdateType.MESSAGE,
    )
    handler.on_registration()
    assert await handler.match(message)


@pytest.mark.asyncio
async def test_match_true(handler_fn, message):
    message.text = "/foo"
    handler = Handler(
        handler_fn,
        update_type=UpdateType.MESSAGE,
        filters=[Command("foo")],
    )
    handler.on_registration()
    assert await handler.match(message)


@pytest.mark.asyncio
async def test_match_true1(handler_fn, message):
    message.text = "/foo"
    handler = Handler(
        handler_fn,
        update_type=UpdateType.MESSAGE,
        filters=[
            Command("foo"),
            User(message.from_.id),
        ]
    )
    handler.on_registration()
    assert await handler.match(message)


@pytest.mark.asyncio
async def test_match_false(handler_fn, message):
    message.text = "/bar"
    handler = Handler(
        handler_fn,
        update_type=UpdateType.MESSAGE,
        filters=[Command("foo")],
    )
    handler.on_registration()
    assert not await handler.match(message)


@pytest.mark.asyncio
async def test_match_false1(handler_fn, message):
    message.text = "/foo"
    handler = Handler(
        handler_fn,
        update_type=UpdateType.MESSAGE,
        filters=[
            Command("bar"),
            User(1),
        ],
    )
    handler.on_registration()
    assert not await handler.match(message)



from __future__ import annotations

__all__ = (
    "Chat",
    "ChatId",
    "private",
    "group",
    "channel",
)

from typing import TYPE_CHECKING
from typing import TypeAlias
from typing import final

from yatbaf.enums import ChatType

from .base import BaseFilter

if TYPE_CHECKING:
    from yatbaf.types import ChatBoostRemoved
    from yatbaf.types import ChatBoostUpdated
    from yatbaf.types import ChatJoinRequest
    from yatbaf.types import ChatMemberUpdated
    from yatbaf.types import Message
    from yatbaf.typing import FilterPriority

Update: TypeAlias = (
    "ChatBoostRemoved "
    "| ChatBoostUpdated"
    "| ChatJoinRequest"
    "| ChatMemberUpdated "
    "| Message "
)


@final
class Chat(BaseFilter):
    """Chat type filter."""

    __slots__ = (
        "types",
        "_priority",
    )

    def __init__(self, *types: ChatType | str, priority: int = 100) -> None:
        if not types:
            raise ValueError("You must pass at least one type.")
        self.types = frozenset([ChatType(t) for t in types])
        self._priority = priority

    @property
    def priority(self) -> FilterPriority:
        return {"chat": (len(self.types), self._priority)}

    async def check(self, update: Update) -> bool:
        return update.chat.type in self.types


private = Chat("private")
group = Chat("group")
channel = Chat("channel")


@final
class ChatId(BaseFilter):
    """Chat id filter."""

    __slots__ = (
        "ids",
        "_priority",
    )

    def __init__(self, *ids: int, priority: int = 150) -> None:
        if not ids:
            raise ValueError("You must pass at least one id.")
        self.ids = frozenset(ids)
        self._priority = priority

    @property
    def priority(self) -> FilterPriority:
        return {"chat": (len(self.ids), self._priority)}

    async def check(self, update: Update) -> bool:
        return update.chat.id in self.ids



import pytest

from yatbaf.handler import Handler
from yatbaf.router import OnMessage


def test_resolve_guards(guard_false, guard_true):
    router = OnMessage(handler_guards=[guard_true])

    @router(guards=[guard_false])
    async def fn(_):
        pass

    handler = router._handlers[0]
    router.on_registration()
    assert handler._guards == [guard_true, guard_false]


def test_resolve_guards_local(guard_false, guard_true):
    router = OnMessage(
        handler_guards=[(guard_true, "local")],
        routers=[nested := OnMessage()],
    )

    @nested(guards=[guard_false])
    async def fn(_):
        pass

    handler = nested._handlers[0]
    router.on_registration()
    assert handler._guards == [guard_false]


@pytest.mark.asyncio
async def test_guard_true(update, handler_fn, guard_true, mock_mark):
    handler = Handler(
        fn=handler_fn,
        update_type=update.event_type,
        guards=[guard_true],
    )
    await handler.handle(update.event)
    mock_mark.assert_called_once_with(update.event)


@pytest.mark.asyncio
async def test_guard_false(update, handler_fn, guard_false, mock_mark):
    handler = Handler(
        fn=handler_fn,
        update_type=update.event_type,
        guards=[guard_false],
    )
    await handler.handle(update.event)
    mock_mark.assert_not_called()



import pytest

from yatbaf.di import Provide
from yatbaf.di import _is_async_callable
from yatbaf.di import _is_async_generator
from yatbaf.di import validate_provider
from yatbaf.exceptions import DependencyError


def sync_gen():
    yield


async def async_gen():
    yield


def sync_fn():
    return


async def async_fn():
    return


class SyncCallable:

    def __call__(self):
        return


class SyncGenerator:

    def __call__(self):
        yield


class AsyncCallable:

    async def __call__(self):
        return


class AsyncGenerator:

    async def __call__(self):
        yield


@pytest.mark.parametrize("obj", [async_fn, AsyncCallable()])
def test_is_async_callable_true(obj):
    assert _is_async_callable(obj)


@pytest.mark.parametrize(
    "obj",
    [
        sync_fn,
        sync_gen,
        async_gen,
        SyncCallable(),
        SyncGenerator(),
        AsyncGenerator(),
    ]
)
def test_is_async_callable_false(obj):
    assert not _is_async_callable(obj)


@pytest.mark.parametrize("obj", [async_gen, AsyncGenerator()])
def test_is_async_generator_true(obj):
    assert _is_async_generator(obj)


@pytest.mark.parametrize(
    "obj", [
        async_fn,
        sync_fn,
        sync_gen,
        SyncCallable(),
        SyncGenerator(),
    ]
)
def test_is_async_generator_false(obj):
    assert not _is_async_generator(obj)


def test_validate_provider_reserved_name():
    provider = Provide(async_fn)
    with pytest.raises(DependencyError):
        validate_provider("update", provider, {})


def test_validate_provider_diff_name():
    with pytest.raises(DependencyError):
        validate_provider(
            "data1", Provide(async_fn), {"data": Provide(async_fn)}
        )


def test_validate_provider_override():
    validate_provider("data", Provide(async_fn), {"data": Provide(async_fn)})
    validate_provider("data", Provide(async_fn), {"data": Provide(async_gen)})



import pytest

from yatbaf.enums import ChatType
from yatbaf.filters import Chat
from yatbaf.filters import ChatId

CHAT_ID = 123232


@pytest.fixture(autouse=True)
def _set_chat_id(chat):
    chat.id = CHAT_ID


def test_priority():
    assert ChatId(123).priority == {"chat": (1, 150)}
    assert ChatId(123, 321).priority == {"chat": (2, 150)}


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "t", [
        ChatType.GROUP,
        ChatType.SUPERGROUP,
        ChatType.PRIVATE,
        ChatType.CHANNEL,
    ]
)
async def test_chat_type(message, t):
    message.chat.type = t
    assert await Chat(t).check(message)


@pytest.mark.asyncio
async def test_chat_type_mix(message):
    message.chat.type = ChatType.GROUP
    assert await Chat("group", "channel").check(message)


@pytest.mark.asyncio
async def test_chat_id(message):
    assert await ChatId(CHAT_ID).check(message)
    assert await ChatId(CHAT_ID, 123).check(message)
